from dataclasses import dataclass
import dataclasses
from pathlib import Path
from typing import Callable, Literal, Mapping, Optional, Sequence, TypeVar, Union, cast
import sqlglot as sg
from odbc2deltalake.destination.destination import (
    Destination,
)
from odbc2deltalake.logging import StorageBackend
from odbc2deltalake.reader import DataSourceReader
from .metadata import (
    get_primary_keys,
    get_columns,
    InformationSchemaColInfo,
)
import sqlglot.expressions as ex
from .sql_glot_utils import table_from_tuple
import logging
from .delta_logger import DeltaLogger


IS_DELETED_COL_NAME = "__is_deleted"
VALID_FROM_COL_NAME = "__timestamp"
IS_FULL_LOAD_COL_NAME = "__is_full_load"


T = TypeVar("T")

_default_type_map = {
    "datetime": ex.DataType(this="datetime2(6)"),
    "datetime2": ex.DataType(this="datetime2(6)"),
    "rowversion": ex.DataType.Type.BIGINT,
    "timestamp": ex.DataType.Type.BIGINT,
    "tinyint": ex.DataType.Type.SMALLINT,  # tinyint is unsigned in T-SQL, therefore we map it to smallint in spark
}
DEFAULT_DATA_TYPE_MAP: Mapping[str, ex.DataType] = _default_type_map


def compat_name(inf: InformationSchemaColInfo) -> str:
    invalid_chars = " ,;{}()\n\t="
    res = inf.column_name
    for ic in invalid_chars:
        res = res.replace(ic, "_")
    return res


class DBDeltaPathConfigs:
    DELTA_1_NAME = "delta_1"
    """common data for delta load contains data that has changed after last full load via naive criteria timestamp > last_timestamp (or similar, can also be dates)"""
    DELTA_2_NAME = "delta_2"
    """delta 2 is data where timestamp is different for whatever reason, eg restore"""
    PRIMARY_KEYS_TS = "primary_keys_ts"
    """file with primary keys and timestamps as of now, before load"""

    LATEST_PK_VERSION = "latest_pk_version"
    """file with primary keys and timestamps as of after load. 
      this will be identital to primary_keys file IF there are no updates in the source within the load. 
      this is unlikely, but possible
    """


@dataclass(frozen=True)
class WriteConfig:
    dialect: str = "tsql"
    """The sqlglot dialect to use for the SQL generation against the source"""

    primary_keys: Union[list[str], None] = None
    """A list of primary keys to use for the delta load. If None, the primary keys will be determined from the source"""

    delta_col: Union[str, None] = None
    """The column to use for the delta load. If None, the column will be determined from the source. Should be mostly increasing to make load efficient"""

    load_mode: Literal[
        "overwrite",
        "append",
        "force_full",
        "append_inserts",
        "simple_delta",
        "simple_delta_check",
    ] = "append"
    """The load mode to use. Attention: overwrite will not help you build scd2, the history is in the delta table only
        append_inserts is for when you have a delta column which is strictly increasing and you want to append new rows only. No deletes of rows. might be good for logs
        simple_delta is for sources where the delta col is a datetime and you can be sure that there are no deletes or additional updates
        simple_delta_check is like simple_delta, but checks for deletes if the count does not match. Only use if you do not expect frequent deletes, as it will do simple_delta AND delta if there are deletes, which is slower than delta
    """

    data_type_map: Mapping[str, ex.DataType] = dataclasses.field(
        default_factory=lambda: _default_type_map.copy()
    )
    """Set this if you want to map stuff like decimal to double before writing to delta. We recommend doing so later in ETL usually"""

    no_complex_entries_load: bool = False
    """If true, will not load 'strange updates' via OPENJSON. Use if your db does not support OPENJSON or you're fine to get some additional updates in order to reduce complexity"""

    get_target_name: Callable[[InformationSchemaColInfo], str] = dataclasses.field(
        default_factory=lambda: compat_name
    )
    """A method that returns the target name of a column. This is used to map the source column names to the target column names.
    Use if you want to apply some naming convention or avoid special characters in the target. """

    allow_schema_drift: Union[bool, Literal["new_only"]] = "new_only"


@dataclass(frozen=True)
class WriteConfigAndInfos:
    col_infos: Sequence[InformationSchemaColInfo]
    pk_cols: Sequence[InformationSchemaColInfo]
    delta_col: Union[InformationSchemaColInfo, None]
    write_config: WriteConfig
    destination: Destination
    source: DataSourceReader
    table_or_query: Union[ex.Query, tuple[str, str], str]
    logger: DeltaLogger

    def execute(self):
        from .db_to_delta import exec_write_db_to_delta

        return exec_write_db_to_delta(self)

    def check_delta_consistency(self, auto_fix=False):
        from .consistency import check_latest_pk

        res = check_latest_pk(self, raise_if_not_consistent=not auto_fix)
        if res and auto_fix:
            self.logger.warning("PK's inconsistent, remove them")
            (
                self.destination / "delta_load" / DBDeltaPathConfigs.LATEST_PK_VERSION
            ).remove(True)
            return res, True
        return res, False

    def exec_source_query(self, query: Union[ex.Query, str]):
        return self.source.source_sql_to_py(query)

    def from_(self, alias: str) -> ex.Select:
        if isinstance(self.table_or_query, ex.Query):
            return sg.from_(self.table_or_query.subquery().as_(alias))
        return sg.from_(table_from_tuple(self.table_or_query, alias))


def get_delta_col(
    cols: Sequence[InformationSchemaColInfo], dialect: str
) -> Union[InformationSchemaColInfo, None]:
    row_start_col: Union[InformationSchemaColInfo, None] = None
    for c in cols:
        if dialect == "tsql" and c.data_type.this in [
            ex.DataType.Type.ROWVERSION,
        ]:
            return c
        if c.generated_always_type_desc == "AS_ROW_START":
            row_start_col = c
        if c.column_name == "__timestamp":
            return c
    return row_start_col


def make_writer(
    source: Union[DataSourceReader, str],
    table_or_query: Union[str, tuple[str, str], ex.Query],
    destination: Union[Destination, Path],
    write_config: Union[WriteConfig, None] = None,
    logger_name: Optional[str] = None,
    *,
    log_file_path: Optional[Destination] = None,
    log_backend: Optional[Union[StorageBackend, Literal["default"]]] = "default",
):
    if write_config is None:
        write_config = WriteConfig()
    if isinstance(destination, Path):
        from .destination.file_system import FileSystemDestination

        destination = cast(Destination, FileSystemDestination(destination))
    if isinstance(source, str):
        from .reader.odbc_reader import ODBCReader

        source = ODBCReader(source)
    cols = get_columns(source, table_or_query, dialect=write_config.dialect)

    if write_config.delta_col:
        delta_col = next(
            (c for c in cols if c.column_name == write_config.delta_col), None
        )
        if delta_col is None:
            delta_col = next(
                (
                    c
                    for c in cols
                    if write_config.get_target_name(c) == write_config.delta_col
                ),
                None,
            )
        if delta_col is None:
            raise ValueError(
                f"Delta column {write_config.delta_col} not found in source"
            )
    else:
        delta_col = get_delta_col(cols, write_config.dialect)

    _pks = write_config.primary_keys
    if _pks is None and (
        isinstance(table_or_query, str) or isinstance(table_or_query, tuple)
    ):
        _pks = get_primary_keys(source, table_or_query, dialect=write_config.dialect)
    elif _pks is None:
        _pks = []
    pk_cols: Sequence[InformationSchemaColInfo] = []
    for pk in _pks:
        pk_col = next(
            (c for c in cols if c.column_name.casefold() == pk.casefold()), None
        )
        if pk_col is None:
            pk_col = next(
                (
                    c
                    for c in cols
                    if write_config.get_target_name(c).casefold() == pk.casefold()
                ),
                None,
            )
        if pk_col is None:
            raise ValueError(f"Primary key {pk} not found in source")
        pk_cols.append(pk_col)
    assert len(_pks) == len(pk_cols), f"Primary keys not found: {_pks}"

    return WriteConfigAndInfos(
        cols,
        pk_cols=pk_cols,
        delta_col=delta_col,
        write_config=write_config,
        destination=destination,
        source=source,
        logger=DeltaLogger(
            log_file_path if log_file_path is not None else destination / "log",
            source,
            logging.getLogger(__name__),
            log_name=logger_name,
            storage_backend=log_backend,
        ),
        table_or_query=table_or_query,
    )



from .write_init import WriteConfigAndInfos, DBDeltaPathConfigs
from .write_utils.restore_pk import create_last_pk_version_view
import sqlglot as sg
import sqlglot.expressions as ex


def check_latest_pk(infos: WriteConfigAndInfos, raise_if_not_consistent=True):
    if not infos.delta_col or not infos.pk_cols:
        raise ValueError("Primary keys and delta column must be defined")
    postfix = "_" + str(abs(hash(str(infos.destination))))
    lpk_view = "lastest_pk_" + postfix
    infos.source.local_register_update_view(
        infos.destination / "delta_load" / DBDeltaPathConfigs.LATEST_PK_VERSION,
        lpk_view,
    )
    _, view_name, success = create_last_pk_version_view(
        infos, view_prefix="v_" + postfix
    )
    d_cols = list(infos.pk_cols) + [infos.delta_col]
    col_names = [infos.write_config.get_target_name(p) for p in d_cols]
    assert success
    assert view_name is not None
    query1 = sg.except_(
        sg.from_(ex.table_(lpk_view, alias="lpk"))
        .select(*[ex.column(c, "lpk", quoted=True) for c in col_names])
        .select(ex.convert("added in persisted data"), append=True),
        sg.from_(ex.table_(view_name, alias="rs"))
        .select(*[ex.column(c, "rs", quoted=True) for c in col_names])
        .select(ex.convert("added in persisted data"), append=True),
    )
    query2 = sg.except_(
        sg.from_(ex.table_(view_name, alias="rs"))
        .select(*[ex.column(c, "rs", quoted=True) for c in col_names])
        .select(ex.convert("missing in persisted data"), append=True),
        sg.from_(ex.table_(lpk_view, alias="lpk"))
        .select(*[ex.column(c, "lpk", quoted=True) for c in col_names])
        .select(ex.convert("missing in persisted data"), append=True),
    )
    result = infos.source.local_execute_sql_to_py(
        ex.union(query1, query2, distinct=False)
    )
    if result:
        print(result)
        if raise_if_not_consistent:
            raise ValueError("Primary keys are not consistent")
    return result



from typing import Union

ODBC_DRIVER: Union[str, None] = None

drivers: Union[list[str], None] = None


def build_connection_string(
    dict_dt: Union[dict, str],
    *,
    odbc: bool = False,
    odbc_driver: Union[str, None] = None,
):
    if isinstance(dict_dt, str) and not odbc:
        return dict_dt
    global drivers
    global ODBC_DRIVER
    if odbc and not odbc_driver and ODBC_DRIVER is None:
        import pyodbc

        drivers = drivers or pyodbc.drivers()
        prio = [
            "ODBC Driver 18 for SQL Server",
            "ODBC Driver 17 for SQL Server",
            "SQL Server",
        ]
        for p in prio:
            if p in drivers:
                ODBC_DRIVER = p
                break
        if ODBC_DRIVER is None:
            ODBC_DRIVER = next(
                (d for d in drivers if "for SQL Server" in d),
                "ODBC Driver 17 for SQL Server",
            )
        assert ODBC_DRIVER is not None
    if isinstance(dict_dt, str):
        assert ODBC_DRIVER is not None
        return "DRIVER=" + (odbc_driver or ODBC_DRIVER) + ";" + dict_dt
    opts = dict_dt | {"DRIVER": (odbc_driver or ODBC_DRIVER)} if odbc else dict_dt
    return ";".join((k + "=" + str(v) for k, v in opts.items()))



from .destination import Destination
from typing import Literal, cast, TYPE_CHECKING

if TYPE_CHECKING:
    import adlfs


class AzureDestination(Destination):
    def __init__(self, container: str, path: str, storage_options: dict):
        from .azure_utils import convert_options

        self.path = path

        self.container = container
        self.storage_options = storage_options
        import adlfs
        import fsspec

        opts = cast(dict[str, str], convert_options(self.storage_options, "fsspec"))
        self.fs = cast(adlfs.AzureBlobFileSystem, fsspec.filesystem("az", **opts))

    def __truediv__(self, other: str):
        return AzureDestination(
            self.container,
            self.path.removesuffix("/") + "/" + other,
            self.storage_options,
        )

    def to_az_path(self):
        return f"az://{self.container}/{self.path}"

    def mkdir(self):
        pass

    def get_fs_path(self) -> "tuple[adlfs.AzureBlobFileSystem, str]":
        return (self.fs, self.to_az_path())

    def upload_str(self, data: str):
        with self.fs.open(self.to_az_path(), "w", encoding="utf-8") as f:
            f.write(data)  # type: ignore

    def modified_time(self):
        fs, path = self.get_fs_path()
        return fs.modified(path)

    def remove(self, recurse: bool = False):
        fs, path = self.get_fs_path()
        fs.rm(path, recursive=recurse)

    def as_path_options(self, flavor: Literal["fsspec", "object_store"]):
        from .azure_utils import convert_options

        return self.to_az_path(), cast(
            dict[str, str], convert_options(self.storage_options, flavor)
        )

    def as_delta_table(self):
        from .azure_utils import convert_options
        from deltalake import DeltaTable

        return DeltaTable(
            self.to_az_path(),
            storage_options=convert_options(self.storage_options, "object_store"),  # type: ignore
        )

    def __str__(self):
        return self.to_az_path()

    @property
    def parent(self):
        return self.__class__(
            self.container,
            "/".join(self.path.removesuffix("/").split("/")[:-1]),
            self.storage_options,
        )

    def exists(self):
        fs, path = self.get_fs_path()
        return cast(bool, fs.exists(path))



from odbc2deltalake.reader.reader import ColInfo
from .reader import DataSourceReader, DeltaOps
from ..destination import Destination
from sqlglot.expressions import Query, DataType
import sqlglot as sg
from typing import Literal, TYPE_CHECKING, Callable, Mapping, Optional, Sequence, Union

if TYPE_CHECKING:
    from pyspark.sql import SparkSession, DataFrame
    from pyspark.sql.types import StructType
    from odbc2deltalake.metadata import InformationSchemaColInfo


class SparkDeltaOps(DeltaOps):
    def __init__(self, dest: Destination, spark):
        from delta.tables import DeltaTable

        self.dest = dest
        self.spark = spark
        self.table = DeltaTable.forPath(spark, str(dest))

    def version(self) -> int:
        return self.table.history(1).select("version").collect()[0].version

    def vacuum(self, retention_hours: Union[int, None] = None):
        self.table.vacuum(retention_hours)

    def restore(self, target: int):
        self.table.restoreToVersion(target)

    def set_properties(self, props: dict[str, str]):
        def _escape(s: str):
            return s.replace("'", "''")

        prop_str = ", ".join(
            [f"'{_escape(k)}' = '{_escape(v)}'" for k, v in props.items()]
        )
        self.spark.sql(
            f"ALTER TABLE delta.`{str(self.dest)}` SET TBLPROPERTIES {prop_str}"
        )

    def get_property(self, key: str) -> Optional[str]:
        from pyspark.sql.functions import col, lit

        res = (
            self.spark.sql(f"show tblproperties delta.`{str(self.dest)}`")
            .where(col("key") == lit(key))
            .select("value")
            .collect()
        )
        if res:
            return res[0].value
        return None

    def columns(self):
        return self.spark.read.format("delta").load(str(self.dest)).columns

    def column_infos(self) -> Sequence[ColInfo]:
        return self.spark.catalog.listColumns("delta.`" + str(self.dest) + "`")

    def set_nullable(self, cols: Mapping[str, bool]):
        for c, n in cols.items():
            if n:
                self.spark.sql(
                    f"ALTER TABLE delta.`{str(self.dest)}` CHANGE COLUMN `{c}` DROP NOT NULL"
                )
            else:
                self.spark.sql(
                    f"ALTER TABLE delta.`{str(self.dest)}` CHANGE COLUMN `{c}` SET NOT NULL"
                )


class SparkReader(DataSourceReader):
    def __init__(
        self,
        spark: "SparkSession",
        sql_config: Optional[dict[str, str]] = None,
        linked_server_proxy: Union[str, None] = None,
        spark_format: str = "sqlserver",
        jdbc=False,
        transformation_hook: Optional[Callable[["DataFrame", str], "DataFrame"]] = None,
    ):
        self.spark = spark
        self.sql_config = sql_config or dict()
        self.linked_server_proxy = linked_server_proxy
        self.spark_format = spark_format
        self.jdbc = jdbc
        self.transformation_hook: Callable[["DataFrame", str], "DataFrame"] = (
            transformation_hook or (lambda d, _: d)
        )
        try:
            self._dialect = (
                "databricks"
                if "/databricks" in (spark.conf.get("spark.home") or "")
                else "spark"
            )
        except Exception:
            self._dialect = "spark"

    def local_register_update_view(
        self,
        delta_path: Destination,
        view_name: str,
        *,
        version: Union[int, None] = None,
    ):
        read = self.spark.read.format("delta")
        if version is not None:
            read = read.option("versionAsOf", version)
        read.load(str(delta_path)).createOrReplaceTempView(view_name)

    def local_register_view(self, sql: Query, view_name: str):
        self.spark.sql(
            f"CREATE OR REPLACE TEMPORARY VIEW {view_name} AS {sql.sql(self._dialect)}"
        )

    def local_execute_sql_to_py(self, sql: Query) -> list[dict]:
        spark_rows = self.spark.sql(sql.sql(self._dialect)).collect()
        return [row.asDict() for row in spark_rows]

    def local_execute_sql_to_delta(
        self,
        sql: Query,
        delta_path: Destination,
        mode: Literal["overwrite", "append"],
        *,
        allow_schema_drift: Union[bool, Literal["new_only"]],
    ):
        df = self.spark.sql(sql.sql(self._dialect))
        writer = df.write.format("delta")
        if allow_schema_drift == "new_only":
            self._append_new_cols(delta_path, df.schema)
        elif allow_schema_drift:
            writer = writer.option(
                "mergeSchema" if mode == "append" else "overwriteSchema", "true"
            )
        writer.mode(mode).save(str(delta_path))

    def local_pylist_to_delta(
        self,
        pylist: list[dict],
        delta_path: Destination,
        mode: Literal["overwrite", "append"],
        dummy_record: Union[dict, None] = None,
    ):
        schema = (
            self.spark.createDataFrame([dummy_record]).schema if dummy_record else None  # type: ignore
        )
        df = self.spark.createDataFrame(pylist, schema=schema)  # type: ignore
        df.write.format("delta").option(
            "mergeSchema" if mode == "append" else "overwriteSchema", "true"
        ).mode(mode).save(str(delta_path))

    @property
    def query_dialect(self) -> str:
        return self._dialect

    @property
    def supports_proc_exec(self):
        return False

    def _query(self, sql: Union[str, Query]):
        if isinstance(sql, Query):
            sql = sql.sql("tsql")
        if self.linked_server_proxy:
            assert "--" not in self.linked_server_proxy
            assert "/*" not in self.linked_server_proxy
            assert "*/" not in self.linked_server_proxy
            sql_escaped = sql.replace("'", "''")
            return f"select * from openquery([{self.linked_server_proxy}], '{ sql_escaped}')"
        return sql

    def source_schema_limit_one(self, sql: Query) -> "list[InformationSchemaColInfo]":
        from ..metadata import InformationSchemaColInfo

        limit_query = sql.limit(0)
        reader = self._reader(limit_query)
        df = self.transformation_hook(reader.load(), "metadata")
        return [
            InformationSchemaColInfo(
                column_name=col.name,
                data_type=DataType.build(col.dataType.simpleString(), dialect="spark"),
                is_nullable=col.nullable,
            )
            for col in df.schema.fields
        ]

    def source_sql_to_py(self, sql: Union[str, Query]) -> list[dict]:
        reader = self._reader(sql)
        rows = self.transformation_hook(reader.load(), "source2py").collect()
        return [row.asDict() for row in rows]

    def local_delta_table_exists(
        self, delta_path: Destination, extended_check=False
    ) -> bool:
        from delta import DeltaTable

        if delta_path.exists() and DeltaTable.isDeltaTable(self.spark, str(delta_path)):
            if extended_check:
                return (
                    len(
                        self.spark.sql(
                            f"select * from delta.`{str(delta_path)}` limit 0"
                        ).columns
                    )
                    > 0
                )
            return True
        else:
            return False

    def _reader(self, sql: Union[str, Query]):
        if self.jdbc:
            options = {}
            jdbcUrl = f"jdbc:{self.spark_format}://"
            if "host" in self.sql_config:
                jdbcUrl += self.sql_config["host"].replace(",", ":")
            if "server" in self.sql_config:
                jdbcUrl += self.sql_config["server"].replace(",", ":")
            if "port" in self.sql_config:
                jdbcUrl += ":" + str(self.sql_config["port"])

            for key, value in self.sql_config.items():
                if key.lower() in ["host", "port", "server"]:
                    continue
                if key.lower() in [
                    "encrypt",
                    "TrustServerCertificate".lower(),
                    "integratedSecurity".lower(),
                ]:
                    enc_vl = value
                    if enc_vl.lower() == "yes":
                        enc_vl = "true"
                    elif enc_vl.lower() == "no":
                        enc_vl = "false"
                    assert enc_vl in ["true", "false"]
                    jdbcUrl += f";{key}=" + enc_vl
                elif key.lower() == "database":
                    jdbcUrl += ";databaseName=" + value
                else:
                    options[key] = value
            print(jdbcUrl)
            reader = self.spark.read.format("jdbc").option("url", jdbcUrl)
        else:
            options = self.sql_config
            reader = self.spark.read.format(self.spark_format)
        reader = reader.option("query", self._query(sql))
        for k, v in options.items():
            reader = reader.option(k, v)
        return reader

    def _append_new_cols(self, delta_path: Destination, source_schema: "StructType"):
        from delta import DeltaTable

        if DeltaTable.isDeltaTable(self.spark, str(delta_path)):
            empty_read_df = (
                self.spark.read.format("delta").load(str(delta_path)).limit(0)
            )
            existing_schema = empty_read_df.schema
            existing_fields = existing_schema.fieldNames()
            existing_fields_lower = [f.lower() for f in existing_fields]
            new_schema = existing_schema
            has_new = False
            for f in source_schema.fields:
                if f.name.lower() not in existing_fields_lower:
                    has_new = True
                    new_schema = new_schema.add(f.name, f.dataType)

            if has_new:
                new_cols = self.spark.createDataFrame([], new_schema)
                new_cols.write.format("delta").option("mergeSchema", "true").mode(
                    "append"
                ).save(str(delta_path))

    def source_write_sql_to_delta(
        self,
        sql: str,
        delta_path: Destination,
        mode: Literal["overwrite", "append"],
        *,
        allow_schema_drift: Union[bool, Literal["new_only"]],
    ):
        reader = self._reader(sql)
        reader = self.transformation_hook(reader.load(), "sql2delta")
        writer = reader.write.format("delta")
        if allow_schema_drift == "new_only":
            self._append_new_cols(delta_path, reader.schema)
        elif allow_schema_drift:
            writer = writer.option(
                "mergeSchema" if mode == "append" else "overwriteSchema", "true"
            )
        writer.mode(mode).save(str(delta_path))

    def get_local_delta_ops(self, delta_path: Destination) -> DeltaOps:
        return SparkDeltaOps(delta_path, self.spark)

    def local_upsert_into(
        self,
        local_sql_source: Query,
        target_delta: Destination,
        merge_cols: Sequence[str],
    ):
        from delta.tables import DeltaTable

        assert len(merge_cols) > 0
        df_source = self.spark.sql(local_sql_source.sql(self._dialect))

        DeltaTable.forPath(self.spark, str(target_delta)).alias("tgt").merge(
            df_source.alias("src"),
            sg.and_(
                *[
                    sg.column(mc, "tgt", quoted=True).eq(
                        sg.column(mc, "src", quoted=True)
                    )
                    for mc in merge_cols
                ]
            ).sql(self._dialect),
        ).whenNotMatchedInsertAll().whenMatchedUpdateAll().execute()



from datetime import datetime, timezone
from odbc2deltalake.destination.destination import Destination
import pydantic
import logging
from uuid import uuid4
from typing import Literal, Optional, Union
from .logging import LogMessage, StorageBackend
from odbc2deltalake.reader.reader import DataSourceReader

is_pydantic_2 = int(pydantic.__version__.split(".")[0]) > 1


class DeltaStorageBackend(StorageBackend):
    _pending_logs: list[LogMessage] = []

    def __init__(self, log_file_path: Destination, source: DataSourceReader):
        self.log_file_path = log_file_path
        self.source = source

    def log(self, msg: LogMessage):
        self._pending_logs.append(msg)
        if len(self._pending_logs) > 10:
            self.flush()

    def flush(self):
        dummy = LogMessage(
            message="",
            type="",
            date=datetime.now(tz=timezone.utc),
            sql="",
            load="",
            sub_load="",
            error_trackback="",
            logger_id="id",
            logger_name="dummy",
        )
        self.source.local_pylist_to_delta(
            [p.model_dump() if is_pydantic_2 else p.dict() for p in self._pending_logs],
            self.log_file_path,
            "append",
            dummy_record=dummy.model_dump() if is_pydantic_2 else dummy.dict(),
        )
        self._pending_logs.clear()


class DeltaLogger:
    def __init__(
        self,
        log_file_path: Destination,
        source: DataSourceReader,
        base_logger: Union[logging.Logger, None] = None,
        print_to_console: bool = False,
        log_name: Optional[str] = None,
        storage_backend: Optional[
            Union[StorageBackend, Literal["default"]]
        ] = "default",
    ):
        self.log_file_path = log_file_path
        self.base_logger = base_logger
        self.source = source
        self.id = uuid4()
        self.print_to_console = print_to_console
        self.log_name = log_name
        if storage_backend == "default":
            self.storage_backend = DeltaStorageBackend(log_file_path, source)
        else:
            self.storage_backend = storage_backend

    def _log(self, msg: LogMessage):
        if self.storage_backend:
            self.storage_backend.log(msg)
        if self.base_logger or self.print_to_console:
            msg_str = msg.message
            if msg.sql:
                msg_str += f" | SQL: {msg.sql}"
            if msg.error_trackback:
                msg_str += f" | Error: {msg.error_trackback}"
            if msg.load:
                msg_str += (
                    f" | Load: {msg.load}{', ' + msg.sub_load if msg.sub_load else '' }"
                )

            if self.base_logger and msg.type == "info":
                self.base_logger.info(msg_str)
            elif self.base_logger and (msg.type == "warn" or msg.type == "warning"):
                self.base_logger.warning(msg_str)
            elif self.base_logger and msg.type == "error":
                self.base_logger.error(msg_str)

            if self.print_to_console:
                print(msg.type + ": " + msg_str)

    def info(
        self,
        message: str,
        *,
        load: Union[str, None] = None,
        sql: Union[str, None] = None,
        sub_load: Union[str, None] = None,
    ):
        self._log(
            LogMessage(
                message=message,
                type="info",
                date=datetime.now(tz=timezone.utc),
                load=load,
                sql=sql,
                sub_load=sub_load,
                logger_id=str(self.id),
                logger_name=self.log_name,
            )
        )

    def warning(
        self,
        message: str,
        *,
        load: Union[str, None] = None,
        sql: Union[str, None] = None,
        sub_load: Union[str, None] = None,
        error_trackback: Union[str, None] = None,
    ):
        self._log(
            LogMessage(
                message=message,
                type="warn",
                date=datetime.now(tz=timezone.utc),
                load=load,
                sql=sql,
                sub_load=sub_load,
                error_trackback=error_trackback,
                logger_id=str(self.id),
                logger_name=self.log_name,
            )
        )

    def error(
        self,
        message: str,
        *,
        load: Union[str, None] = None,
        sql: Union[str, None] = None,
        sub_load: Union[str, None] = None,
        error_trackback: Union[str, None] = None,
    ):
        self._log(
            LogMessage(
                message=message,
                type="error",
                date=datetime.now(tz=timezone.utc),
                load=load,
                sql=sql,
                sub_load=sub_load,
                error_trackback=error_trackback,
                logger_id=str(self.id),
                logger_name=self.log_name,
            )
        )

    def flush(self):
        if self.storage_backend:
            self.storage_backend.flush()



from .write_init import (
    make_writer,
    DBDeltaPathConfigs,
    WriteConfig,
    DEFAULT_DATA_TYPE_MAP,
    WriteConfigAndInfos,
)
from .reader.reader import DataSourceReader
from .destination.destination import Destination
from pathlib import Path
from typing import Union


def write_db_to_delta(
    source: Union[DataSourceReader, str],
    table_or_query: tuple[str, str],
    destination: Union[Destination, Path],
    write_config: Union[WriteConfig, None] = None,
):
    make_writer(
        source=source,
        table_or_query=table_or_query,
        destination=destination,
        write_config=write_config,
    ).execute()



from pathlib import Path

from odbc2deltalake.destination.destination import Destination
from odbc2deltalake.reader.reader import ColInfo, DeltaOps
from .reader import DataSourceReader
from typing import TYPE_CHECKING, Literal, Mapping, Optional, Sequence, Union
from sqlglot.expressions import Query, DataType
import sqlglot as sg

if TYPE_CHECKING:
    import pyarrow as pa
    from odbc2deltalake.metadata import InformationSchemaColInfo
    from deltalake import DeltaTable


def _all_nullable(schema: "pa.Schema") -> "pa.Schema":
    sc = schema
    for i, n in enumerate(schema.names):
        f = schema.field(n)
        if not f.nullable:
            sc = sc.set(i, f.with_nullable(True))
    return sc


def _get_type(tp: "pa.DataType"):
    import pyarrow.types as pat

    if pat.is_string(tp):
        return DataType.Type.NVARCHAR
    if pat.is_boolean(tp):
        return DataType.Type.BIT
    if pat.is_int8(tp):
        return DataType.Type.TINYINT
    if pat.is_int16(tp):
        return DataType.Type.SMALLINT
    if pat.is_int32(tp):
        return DataType.Type.INT
    if pat.is_int64(tp):
        return DataType.Type.BIGINT
    if pat.is_float32(tp):
        return DataType.Type.FLOAT
    if pat.is_float64(tp):
        return DataType.Type.DOUBLE
    if pat.is_date32(tp):
        return DataType.Type.DATE
    if pat.is_date64(tp):
        return DataType.Type.DATETIME
    if pat.is_timestamp(tp):
        return DataType.Type.DATETIME
    if pat.is_time32(tp):
        return DataType.Type.TIME
    if pat.is_time64(tp):
        return DataType.Type.TIME
    if pat.is_decimal(tp):
        return DataType.Type.DECIMAL
    if pat.is_binary(tp):
        return DataType.Type.VARBINARY
    if pat.is_fixed_size_binary(tp):
        return DataType.build(f"binary({tp.byte_width})", dialect="tsql")
    raise ValueError(f"Type {tp} not supported")


def _build_type(t: Union[DataType, DataType.Type]):
    if isinstance(t, DataType):
        return t
    return DataType(this=t)


class DeltaRSDeltaOps(DeltaOps):
    def __init__(self, delta_table: "DeltaTable"):
        self.delta_table = delta_table

    def version(self) -> int:
        return self.delta_table.version()

    def vacuum(self, retention_hours: Union[int, None] = None):
        self.delta_table.vacuum(retention_hours)

    def restore(self, target: int):
        self.delta_table.restore(target)

    def set_properties(self, props: dict[str, str]):
        self.delta_table.delete("1=0", custom_metadata=props)

    def get_property(self, key: str):
        return self.delta_table.metadata().configuration.get(key, None)

    def column_infos(self) -> Sequence[ColInfo]:
        return self.delta_table.schema().fields

    def set_nullable(self, cols: Mapping[str, bool]):
        # not implemented properly for delta-rs
        for k, v in cols.items():
            if v:
                self.delta_table.alter.add_constraint({f"null_{k}": f"{k} IS NOT NULL"})
            else:
                self.delta_table.alter.drop_constraint(
                    f"null_{k}", raise_if_not_exists=False
                )


class ODBCReader(DataSourceReader):
    def __init__(self, connection_string: str, local_db: str = ":memory:") -> None:
        from deltalake import WriterProperties

        self.connection_string = connection_string
        self.writer_properties = WriterProperties(compression="ZSTD")
        self.duck_con = None
        self.local_db = local_db

    @property
    def supports_proc_exec(self):
        return True

    def local_register_update_view(
        self,
        delta_path: Destination,
        view_name: str,
        *,
        version: Union[int, None] = None,
    ):
        import duckdb
        from deltalake2db import duckdb_create_view_for_delta

        self.duck_con = self.duck_con or duckdb.connect(self.local_db)
        dt = delta_path.as_delta_table()
        if version is not None:
            dt.load_as_version(version)
        duckdb_create_view_for_delta(self.duck_con, dt, view_name)

    def local_execute_sql_to_py(self, sql: Query) -> list[dict]:
        import duckdb

        self.duck_con = self.duck_con or duckdb.connect(self.local_db)
        with self.duck_con.cursor() as cursor:
            cursor.execute(sql.sql(self.query_dialect))
            assert cursor.description is not None
            col_names = [desc[0] for desc in cursor.description]
            return [dict(zip(col_names, row)) for row in cursor.fetchall()]

    def local_register_view(self, sql: Query, view_name: str):
        import duckdb

        self.duck_con = self.duck_con or duckdb.connect(self.local_db)
        self.duck_con.sql(
            f"CREATE OR REPLACE VIEW {view_name} AS {sql.sql(self.query_dialect)}"
        )

    def local_pylist_to_delta(
        self,
        pylist: list[dict],
        delta_path: Destination,
        mode: Literal["overwrite", "append"],
        dummy_record: Union[dict, None] = None,
    ):
        from deltalake import write_deltalake
        import pyarrow as pa

        schema = (
            pa.RecordBatch.from_pylist([dummy_record]).schema if dummy_record else None
        )
        batch = pa.RecordBatch.from_pylist(pylist, schema=schema)
        dp, do = delta_path.as_path_options("object_store")

        write_deltalake(
            dp,
            [batch],
            storage_options=do,
            schema=batch.schema,
            mode=mode,
            schema_mode="overwrite" if mode == "overwrite" else "merge",
            writer_properties=self.writer_properties,
            engine="rust",
        )

    def local_delta_table_exists(
        self, delta_path: Destination, extended_check=False
    ) -> bool:
        if delta_path.exists():
            from deltalake import DeltaTable
            from deltalake.exceptions import TableNotFoundError

            try:
                dt = delta_path.as_delta_table()
                DeltaTable.version(dt)
                if extended_check:
                    return len(dt.schema().fields) > 0
                return True
            except TableNotFoundError:
                return False
        return False

    def _handle_schema_drift(
        self,
        delta_path: Destination,
        allow_schema_drift: Union[bool, Literal["new_only"]],
        mode: Literal["overwrite", "append"],
        input_schema,
    ):
        from deltalake.exceptions import TableNotFoundError

        dp, do = delta_path.as_path_options("object_store")
        schema_mode = None
        if allow_schema_drift == "new_only":
            try:
                existing_schema = delta_path.as_delta_table().schema().fields
                existing_field_names = [f.name for f in existing_schema]
                new_schema = _all_nullable(input_schema)
                new_fields = [
                    new_schema.field(n)
                    for n in new_schema.names
                    if n not in existing_field_names
                ]

                if new_fields:
                    import pyarrow as pa

                    merge_schema_fields = list()

                    pas_schema = delta_path.as_delta_table().schema().to_pyarrow()
                    for n in input_schema.names:
                        if n in existing_field_names:
                            merge_schema_fields.append(pas_schema.field(n))
                        else:
                            merge_schema_fields.append(input_schema.field(n))

                    return pa.schema(merge_schema_fields), "merge"
            except TableNotFoundError:
                pass
        elif allow_schema_drift:
            schema_mode = "overwrite" if mode == "overwrite" else "merge"
        return None, schema_mode

    def local_execute_sql_to_delta(
        self,
        sql: Query,
        delta_path: Destination,
        mode: Literal["overwrite", "append"],
        *,
        allow_schema_drift: Union[bool, Literal["new_only"]],
    ):
        import duckdb
        from deltalake import write_deltalake
        from deltalake.exceptions import DeltaError

        self.duck_con = self.duck_con or duckdb.connect(self.local_db)

        with self.duck_con.cursor() as cur:
            cur.execute(sql.sql(self.query_dialect))
            dp, do = delta_path.as_path_options("object_store")
            batch_reader = cur.fetch_record_batch()
            schema = batch_reader.schema
            cast_schema, schema_mode = self._handle_schema_drift(
                delta_path, allow_schema_drift, mode, schema
            )
            if cast_schema:
                schema = cast_schema
                batch_reader = pa.RecordBatchReader.from_batches(
                    cast_schema, (b.cast(cast_schema) for b in batch_reader)
                )
            try:
                write_deltalake(
                    dp,
                    batch_reader,
                    mode=mode,
                    schema_mode=schema_mode,
                    writer_properties=self.writer_properties,
                    engine="rust",
                    storage_options=do,
                )
            except DeltaError as e:
                if "No data source supplied to write command" in str(e):
                    if mode == "overwrite":
                        self._write_empty_delta_table(schema, dp, do)
                else:
                    raise e

    @property
    def query_dialect(self) -> str:
        return "duckdb"

    def source_schema_limit_one(self, sql: Query) -> "list[InformationSchemaColInfo]":
        from ..metadata import InformationSchemaColInfo

        limit_sql = sql.limit(0).sql("tsql")
        from arrow_odbc import read_arrow_batches_from_odbc

        sc = read_arrow_batches_from_odbc(
            limit_sql, self.connection_string, max_text_size=20000
        ).schema
        return [
            InformationSchemaColInfo(
                column_name=n,
                data_type=_build_type(_get_type(sc.field(n).type)),
                is_nullable=sc.field(n).nullable,
            )
            for n in sc.names
        ]

    def source_sql_to_py(self, sql: Union[str, Query]) -> list[dict]:
        if isinstance(sql, Query):
            sql = sql.sql("tsql")
        from arrow_odbc import read_arrow_batches_from_odbc

        result = list()
        for batch in read_arrow_batches_from_odbc(
            sql, self.connection_string, max_text_size=20000
        ):
            result.extend(batch.to_pylist())
        return result

    def source_write_sql_to_delta(
        self,
        sql: str,
        delta_path: Destination,
        mode: Literal["overwrite", "append"],
        *,
        allow_schema_drift: Union[bool, Literal["new_only"]],
    ):
        from arrow_odbc import read_arrow_batches_from_odbc
        from deltalake import write_deltalake
        from deltalake.exceptions import DeltaError

        reader = read_arrow_batches_from_odbc(
            query=sql,
            connection_string=self.connection_string,
            max_binary_size=20000,
            max_text_size=20000,
        )
        dp, do = delta_path.as_path_options(flavor="object_store")
        cast_schema, schema_mode = self._handle_schema_drift(
            delta_path, allow_schema_drift, mode, reader.schema
        )

        if cast_schema:
            import pyarrow as pa

            reader = pa.RecordBatchReader.from_batches(
                cast_schema, (b.cast(cast_schema) for b in reader)
            )
        try:
            write_deltalake(
                dp,
                reader,
                schema=_all_nullable(reader.schema),
                mode=mode,
                writer_properties=self.writer_properties,
                schema_mode=schema_mode,
                engine="rust",
                storage_options=do,
            )
        except DeltaError as e:
            if "No data source supplied to write command" in str(e):
                if mode == "overwrite":
                    self._write_empty_delta_table(reader.schema, dp, do)
            else:
                raise e

    def _write_empty_delta_table(
        self,
        schema: "pa.Schema",
        path: Union[str, Path],
        storage_options: Optional[dict[str, str]],
    ):
        from deltalake import write_deltalake

        write_deltalake(
            path,
            [],
            schema=_all_nullable(schema),
            mode="overwrite",
            schema_mode="overwrite",
            engine="pyarrow",
            storage_options=storage_options,
        )

    def get_local_delta_ops(self, delta_path: Destination) -> DeltaOps:
        dt = delta_path.as_delta_table()
        return DeltaRSDeltaOps(dt)

    def local_upsert_into(
        self,
        local_sql_source: Query,
        target_delta: Destination,
        merge_cols: Sequence[str],
    ):
        import duckdb

        self.duck_con = self.duck_con or duckdb.connect(self.local_db)

        with self.duck_con.cursor() as cursor:
            cursor.execute(local_sql_source.sql(self.query_dialect))
            dt = target_delta.as_delta_table()
            res = (
                dt.merge(
                    cursor.fetch_record_batch(),
                    sg.and_(
                        *[
                            sg.column(mc, "tgt", quoted=True).eq(
                                sg.column(mc, "src", quoted=True)
                            )
                            for mc in merge_cols
                        ]
                    ).sql(self.query_dialect),
                    "src",
                    "tgt",
                    writer_properties=self.writer_properties,
                )
                .when_matched_update_all()
                .when_not_matched_insert_all()
                .execute()
            )
            print(res)



from typing import Callable, Literal, Optional, Union


default_azure_args = [
    "authority",
    "exclude_workload_identity_credential",
    "exclude_developer_cli_credential",
    "exclude_cli_credential",
    "exclude_environment_credential",
    "exclude_managed_identity_credential",
    "exclude_powershell_credential",
    "exclude_visual_studio_code_credential",
    "exclude_shared_token_cache_credential",
    "exclude_interactive_browser_credential",
    "interactive_browser_tenant_id",
    "managed_identity_client_id",
    "workload_identity_client_id",
    "workload_identity_tenant_id",
    "interactive_browser_client_id",
    "shared_cache_username",
    "shared_cache_tenant_id",
    "visual_studio_code_tenant_id",
    "process_timeout",
]


_token_state = dict()


def _get_default_token(**kwargs) -> str:
    global _token_state
    from azure.identity import DefaultAzureCredential

    cred: Union[DefaultAzureCredential, None] = _token_state.get("cred", None)
    if not cred:
        cred = DefaultAzureCredential(**kwargs)
        _token_state["cred"] = cred
    return cred.get_token("https://storage.azure.com/.default").token


def convert_options(
    options: Union[dict, None],
    flavor: Literal["fsspec", "object_store"],
    token_retrieval_func: Optional[Callable[[], str]] = None,
):
    if options is None:
        return None
    use_emulator: bool = options.get("use_emulator", "0").lower() in ["1", "true"]
    if flavor == "fsspec" and "connection_string" not in options and use_emulator:
        constr = "DefaultEndpointsProtocol=http;AccountName=devstoreaccount1;AccountKey=Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw==;BlobEndpoint=http://127.0.0.1:10000/devstoreaccount1;QueueEndpoint=http://127.0.0.1:10001/devstoreaccount1;TableEndpoint=http://127.0.0.1:10002/devstoreaccount1;"
        return {"connection_string": constr}
    elif (
        flavor == "fsspec"
        and "account_key" not in options
        and "anon" not in options
        and "sas_token" not in options
        and "token" not in options
        and "account_name" in options
    ):  # anon is true by default in fsspec which makes no sense mostly
        return {"anon": False} | options

    new_opts = options.copy()
    anon_value = False
    if flavor == "object_store" and "anon" in options:
        anon_value = new_opts.pop("anon")
    if (
        flavor == "object_store"
        and "account_key" not in options
        and "sas_token" not in options
        and "token" not in options
        and not use_emulator
        and not anon_value
    ):
        token_kwargs = {k: new_opts.pop(k) for k in default_azure_args if k in new_opts}
        new_opts["token"] = (token_retrieval_func or _get_default_token)(**token_kwargs)
    return new_opts



from typing import Literal, TYPE_CHECKING, Union
from .destination import Destination
from pathlib import Path

if TYPE_CHECKING:
    import fsspec


class FileSystemDestination(Destination):
    def __init__(self, path: Union[str, Path]):
        self.path = Path(path)
        import fsspec

        self.fs = fsspec.filesystem("file")

    def mkdir(self):
        self.path.mkdir(parents=True, exist_ok=True)

    def get_fs_path(self) -> "tuple[fsspec.AbstractFileSystem, str]":
        return (self.fs, str(self.path))

    def __str__(self):
        return str(self.path)

    def exists(self):
        return self.path.exists()

    def upload_str(self, data: str):
        with open(self.path, "w", encoding="utf-8") as f:
            f.write(data)

    def modified_time(self):
        fs, path = self.get_fs_path()
        return fs.modified(path)

    def remove(self, recurse: bool = False):
        if recurse:
            import shutil

            shutil.rmtree(self.path)
        else:
            self.path.unlink()

    @property
    def parent(self):
        return self.__class__(self.path.parent)

    def as_path_options(self, flavor: Literal["fsspec", "object_store"]):
        return str(self.path), None

    def as_delta_table(self):
        from deltalake import DeltaTable

        return DeltaTable(self.path)

    def __truediv__(self, other: str):
        return FileSystemDestination(self.path / other)



from odbc2deltalake.write_init import (
    IS_DELETED_COL_NAME,
    IS_FULL_LOAD_COL_NAME,
    VALID_FROM_COL_NAME,
    DBDeltaPathConfigs,
)
from odbc2deltalake.write_init import WriteConfigAndInfos
import sqlglot.expressions as ex
import sqlglot as sg
from typing import Union
from odbc2deltalake.sql_glot_utils import count_limit_one

table_name_type = Union[str, tuple[str, str], tuple[str, str, str]]


def create_last_pk_version_view(
    infos: WriteConfigAndInfos,
    view_prefix: str = "",
):
    assert len(infos.pk_cols) > 0, "must have at least one pk column"
    delta_path = infos.destination / "delta"
    reader = infos.source
    write_config = infos.write_config

    temp_table = "tmp_" + str(abs(hash(str(delta_path))))
    reader.local_register_update_view(delta_path, temp_table)

    sq_valid_from = reader.local_execute_sql_to_py(
        sg.from_(ex.to_identifier(temp_table))
        .select(
            ex.func("max", ex.column(VALID_FROM_COL_NAME, quoted=True)).as_(
                VALID_FROM_COL_NAME, quoted=True
            )
        )
        .where(ex.column(IS_FULL_LOAD_COL_NAME, quoted=True).eq(True))
    )
    if sq_valid_from is None or len(sq_valid_from) == 0:
        return None, None, False
    assert infos.delta_col is not None, "must have a delta column"
    latest_full_load_date = sq_valid_from[0][VALID_FROM_COL_NAME]
    reader.local_register_view(
        sg.from_(ex.table_(ex.to_identifier(temp_table), alias="tr"))
        .select(
            *(
                [
                    ex.column(write_config.get_target_name(c), quoted=True)
                    for c in infos.pk_cols
                ]
                + [
                    ex.column(
                        write_config.get_target_name(infos.delta_col), quoted=True
                    ),
                    ex.column(VALID_FROM_COL_NAME, quoted=True),
                ]
            ),
            copy=False,
        )
        .where(
            ex.column(IS_FULL_LOAD_COL_NAME, quoted=True).eq(True)
            and ex.column(VALID_FROM_COL_NAME, quoted=True).eq(
                ex.Subquery(
                    this=ex.select(
                        ex.func(
                            "MAX", ex.column(VALID_FROM_COL_NAME, "ts", quoted=True)
                        )
                    )
                    .from_(ex.table_(ex.to_identifier(temp_table), alias="ts"))
                    .where(ex.column(IS_FULL_LOAD_COL_NAME, quoted=True).eq(True))
                )
            )
        ),
        view_prefix + "last_full_load",
    )

    sq = (
        sg.from_(ex.table_(temp_table, alias="tr"))
        .select(
            *(
                [
                    ex.column(write_config.get_target_name(c), "tr", quoted=True)
                    for c in infos.pk_cols
                ]
                + [
                    ex.column(
                        write_config.get_target_name(infos.delta_col), "tr", quoted=True
                    )
                ]
                + [ex.column(IS_DELETED_COL_NAME, "tr", quoted=True)]
                + [ex.column(VALID_FROM_COL_NAME, "tr", quoted=True)]
            )
        )
        .where(
            ex.column(VALID_FROM_COL_NAME, quoted=True)
            > ex.convert(latest_full_load_date)
        )
    )
    sq = sq.qualify(
        ex.EQ(
            this=ex.Window(
                this=ex.RowNumber(),
                partition_by=[
                    ex.column(write_config.get_target_name(pk), quoted=True)
                    for pk in infos.pk_cols
                ],
                order=ex.Order(
                    expressions=[
                        ex.Ordered(
                            this=ex.column(VALID_FROM_COL_NAME, quoted=True),
                            desc=True,
                            nulls_first=False,
                        )
                    ]
                ),
                over="OVER",
            ),
            expression=ex.convert(1),
        )
    )
    reader.local_register_view(sq, view_prefix + "delta_after_full_load")
    last_pk_query = (
        sg.from_(ex.table_("base", alias="b"))
        .where(~ex.column(IS_DELETED_COL_NAME, quoted=True, table="b"))
        .with_(
            "base",
            as_=ex.union(
                sg.from_(
                    ex.table_(view_prefix + "delta_after_full_load", alias="df")
                ).select(
                    *(
                        [
                            ex.column(
                                write_config.get_target_name(c), "df", quoted=True
                            )
                            for c in infos.pk_cols
                        ]
                        + [
                            ex.column(
                                write_config.get_target_name(infos.delta_col),
                                "df",
                                quoted=True,
                            ),
                            ex.column(IS_DELETED_COL_NAME, "df", quoted=True),
                        ]
                    )
                ),
                sg.from_(ex.table_(view_prefix + "last_full_load", alias="f"))
                .select(
                    *(
                        [
                            ex.column(write_config.get_target_name(c), "f", quoted=True)
                            for c in infos.pk_cols
                        ]
                        + [
                            ex.column(
                                write_config.get_target_name(infos.delta_col),
                                "f",
                                quoted=True,
                            ),
                            ex.convert(False).as_(IS_DELETED_COL_NAME, quoted=True),
                        ]
                    )
                )
                .join(
                    ex.table_(view_prefix + "delta_after_full_load", alias="d"),
                    join_type="anti",
                    on=ex.and_(
                        *[
                            ex.column(
                                write_config.get_target_name(c), "f", quoted=True
                            ).eq(
                                ex.column(
                                    write_config.get_target_name(c), "d", quoted=True
                                )
                            )
                            for c in infos.pk_cols
                        ]
                    ),
                ),
                distinct=False,
            ),
        )
        .select(
            *(
                [
                    ex.column(write_config.get_target_name(c), "b", quoted=True)
                    for c in infos.pk_cols
                ]
                + [
                    ex.column(
                        write_config.get_target_name(infos.delta_col),
                        "b",
                        quoted=True,
                    )
                ]
            ),
            append=False,
        )
    )
    reader.local_register_view(
        last_pk_query,
        view_prefix + "last_pk_version",
    )
    return last_pk_query, view_prefix + "last_pk_version", True


def restore_last_pk(infos: WriteConfigAndInfos):
    query, view_name, success = create_last_pk_version_view(
        infos=infos,
        view_prefix="v_odbc_load_",
    )
    if not success:
        return False
    assert query is not None
    assert view_name is not None
    infos.logger.info(
        "Restoring last pk version", sql=query.sql(infos.source.query_dialect)
    )

    cnt = infos.source.local_execute_sql_to_py(count_limit_one(view_name))[0]["cnt"]
    if cnt == 0:
        return False
    infos.source.local_execute_sql_to_delta(
        sg.from_(view_name).select(ex.Star()),
        infos.destination / "delta_load" / DBDeltaPathConfigs.LATEST_PK_VERSION,
        mode="overwrite",
        allow_schema_drift=True,
    )
    return True



from .destination import Destination
from typing import Literal
from datetime import datetime, timezone


class DatabricksDestination(Destination):
    def __init__(
        self, dbutils, container: str, path: str, account: str, scheme: str = "abfss"
    ):
        self.path = path.removeprefix("/")
        if "." not in account:
            account = account + ".dfs.core.windows.net"
        self.account = account
        self.container = container
        self.dbutils = dbutils
        self.scheme = scheme
        self.fs = None

    def __truediv__(self, other: str):
        return DatabricksDestination(
            self.dbutils,
            container=self.container,
            path=self.path.removesuffix("/") + "/" + other,
            account=self.account,
            scheme=self.scheme,
        )

    def to_az_path(self):
        return f"{self.scheme}://{self.container}@{self.account}/{self.path}"

    def mkdir(self):
        self.dbutils.fs.mkdirs(self.to_az_path())

    def upload_str(self, data: str):
        self.dbutils.fs.put(self.to_az_path(), data, overwrite=True)

    def modified_time(self):
        res = self.dbutils.fs.ls(self.to_az_path())
        assert len(res) == 1
        mod_time: int = res[0].modificationTime
        return datetime.fromtimestamp(mod_time / 1000, tz=timezone.utc)

    def remove(self, recurse: bool = False):
        self.dbutils.fs.rm(self.to_az_path(), recurse=recurse)

    def as_path_options(self, flavor: Literal["fsspec", "object_store"]):
        raise NotImplementedError("not an option for databricks")

    def as_delta_table(self):
        raise NotImplementedError("not an option for databricks")

    def __str__(self):
        return self.to_az_path()

    @property
    def parent(self):
        new_path = "/".join(self.path.removesuffix("/").split("/")[:-1])

        return DatabricksDestination(
            self.dbutils,
            container=self.container,
            path=new_path,
            account=self.account,
            scheme=self.scheme,
        )

    def exists(self):
        try:
            self.dbutils.fs.ls(self.to_az_path())
            return True
        except Exception:
            return False



from odbc2deltalake.destination.destination import Destination
import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Literal, Mapping, Protocol, Any, Sequence, Union
from sqlglot.expressions import Query

if TYPE_CHECKING:
    from odbc2deltalake.metadata import InformationSchemaColInfo

logger = logging.getLogger(__name__)


class ColInfo(Protocol):
    name: str
    nullable: bool


class DeltaOps(Protocol):
    def version(
        self,
    ) -> int: ...

    def vacuum(self, retention_hours: Union[int, None] = None) -> Any: ...

    def restore(self, target: int) -> Any: ...

    def set_properties(self, props: dict[str, str]): ...

    def get_property(self, key: str) -> Union[str, None]: ...

    def column_infos(self) -> Sequence[ColInfo]: ...

    def set_nullable(self, cols: Mapping[str, bool]): ...


class DataSourceReader(ABC):
    @property
    @abstractmethod
    def supports_proc_exec(self) -> bool:
        pass

    @property
    @abstractmethod
    def query_dialect(self) -> str:
        pass

    @abstractmethod
    def local_delta_table_exists(
        self, delta_path: Destination, extended_check=False
    ) -> bool:
        pass

    @abstractmethod
    def source_write_sql_to_delta(
        self,
        sql: str,
        delta_path: Destination,
        mode: Literal["overwrite", "append"],
        *,
        allow_schema_drift: Union[bool, Literal["new_only"]],
    ):
        pass

    @abstractmethod
    def source_schema_limit_one(self, sql: Query) -> "list[InformationSchemaColInfo]":
        pass

    @abstractmethod
    def source_sql_to_py(self, sql: Union[str, Query]) -> list[dict]:
        pass

    @abstractmethod
    def local_execute_sql_to_py(self, sql: Query) -> list[dict]:
        pass

    @abstractmethod
    def get_local_delta_ops(self, delta_path: Destination) -> DeltaOps:
        pass

    @abstractmethod
    def local_execute_sql_to_delta(
        self,
        sql: Query,
        delta_path: Destination,
        mode: Literal["overwrite", "append"],
        *,
        allow_schema_drift: Union[bool, Literal["new_only"]],
    ):
        pass

    @abstractmethod
    def local_pylist_to_delta(
        self,
        pylist: list[dict],
        delta_path: Destination,
        mode: Literal["overwrite", "append"],
        dummy_record: Union[dict, None] = None,
    ):
        pass

    @abstractmethod
    def local_register_view(self, sql: Query, view_name: str):
        pass

    @abstractmethod
    def local_register_update_view(
        self,
        delta_path: Destination,
        view_name: str,
        *,
        version: Union[int, None] = None,
    ):
        pass

    @abstractmethod
    def local_upsert_into(
        self,
        local_sql_source: Query,
        target_delta: Destination,
        merge_cols: Sequence[str],
    ):
        pass



from typing import Sequence, Union
import sqlglot.expressions as ex
import sqlglot as sg


def union(selects: Sequence[ex.Query], *, distinct: bool) -> ex.Query:
    if len(selects) == 0:
        raise ValueError("No selects to union")
    elif len(selects) == 1:
        return selects[0]
    elif len(selects) == 2:
        return ex.union(selects[0], selects[1], distinct=distinct)
    else:
        return ex.union(
            selects[0], union(selects[1:], distinct=distinct), distinct=distinct
        )


def count_limit_one(table_name: Union[ex.Expression, str]):
    return sg.from_(sg.from_(table_name).select(ex.Star()).limit(1).subquery()).select(
        ex.Count(this=ex.Star()).as_("cnt")
    )


def table_from_tuple(
    name: Union[str, tuple[str, str], tuple[str, str, str]],
    alias: Union[str, None] = None,
) -> ex.Table:
    if alias is not None:
        assert " " not in alias
        assert "-" not in alias
        assert "'" not in alias
        assert '"' not in alias
        assert "*" not in alias
        assert "/" not in alias
        assert "\\" not in alias

    if isinstance(name, str):
        return ex.Table(
            this=ex.Identifier(this=name, quoted=True),
            alias=ex.Identifier(this=alias, quoted=False) if alias else None,
        )
    if len(name) == 2:
        return ex.Table(
            this=ex.Identifier(this=name[1], quoted=True),
            db=ex.Identifier(this=name[0], quoted=True),
            alias=ex.Identifier(this=alias, quoted=False) if alias else None,
        )
    if len(name) == 3:
        return ex.Table(
            this=ex.Identifier(this=name[2], quoted=True),
            db=ex.Identifier(this=name[1], quoted=True),
            catalog=ex.Identifier(this=name[0], quoted=True),
            alias=ex.Identifier(this=alias, quoted=False) if alias else None,
        )
    raise ValueError(f"Invalid name: {name}")



from typing import Literal, Union, Any, Optional


class FullLoadResult:
    executed_type: Literal["full"]

    def __init__(self):
        self.executed_type = "full"


State = tuple[Any, int]


class DeltaLoadResult:
    executed_type: Literal["delta"]
    starting_local_state: State
    starting_source_state: State
    end_source_state: State

    dirty: Optional[bool] = None  # True if the count does not match

    def __init__(self):
        self.executed_type = "delta"


class NoLoadResult:
    executed_type: Literal["none"]

    def __init__(self):
        self.executed_type = "none"


class AppendOnlyLoadResult:
    executed_type: Literal["append_only"]

    def __init__(self):
        self.executed_type = "append_only"


LoadResult = Union[FullLoadResult, DeltaLoadResult, NoLoadResult, AppendOnlyLoadResult]



from typing import Any, TYPE_CHECKING
import sqlglot as sg
import sqlglot.expressions as ex

if TYPE_CHECKING:
    from .write_init import WriteConfigAndInfos


def get_local_delta_value_and_count(
    infos: "WriteConfigAndInfos",
) -> tuple[Any, int]:
    from odbc2deltalake.write_init import DBDeltaPathConfigs

    if (infos.destination / "delta_load" / DBDeltaPathConfigs.PRIMARY_KEYS_TS).exists():
        delta_path = (
            infos.destination / "delta_load" / DBDeltaPathConfigs.PRIMARY_KEYS_TS
        )
    else:
        delta_path = infos.destination / "delta"
    tmp_view_name = "temp_" + str(abs(hash(str(delta_path))))
    infos.source.local_register_update_view(delta_path, tmp_view_name)
    row = infos.source.local_execute_sql_to_py(
        sg.from_(ex.to_identifier(tmp_view_name)).select(
            (
                ex.func(
                    "MAX",
                    ex.column(infos.write_config.get_target_name(infos.delta_col)),
                ).as_("max_ts")
                if infos.delta_col
                else ex.convert(None).as_("max_ts")
            ),
            ex.Count(this=ex.Star()).as_("cnt"),
        )
    )[0]
    mt, cnt = row["max_ts"], row["cnt"]
    if isinstance(mt, bytearray):
        return bytes(mt), cnt
    return mt, cnt


def retrieve_source_ts_cnt(infos: "WriteConfigAndInfos"):
    pk_ts_col_select = infos.from_("t").select(
        (
            ex.func(
                "MAX",
                ex.column(infos.delta_col.column_name, quoted=True),
            ).as_("max_ts")
            if infos.delta_col
            else ex.convert(None).as_("max_ts")
        ),
        ex.Count(this=ex.Star()).as_("cnt"),
    )

    infos.logger.info(
        "Retrieve source delta value / count",
        sql=pk_ts_col_select.sql(infos.write_config.dialect),
    )
    row = infos.source.source_sql_to_py(sql=pk_ts_col_select)
    return row[0]["max_ts"], row[0]["cnt"]



from datetime import datetime
from pydantic import BaseModel
from typing import Optional, Union, Protocol


class LogMessage(BaseModel):
    message: str
    type: str
    date: datetime
    logger_id: str
    logger_name: Optional[str]
    sql: Union[str, None] = None
    load: Union[str, None] = None
    sub_load: Union[str, None] = None
    error_trackback: Union[str, None] = None


class StorageBackend(Protocol):
    def log(self, msg: LogMessage): ...

    def flush(self): ...



import dataclasses
from datetime import datetime, timezone
from typing import (
    Callable,
    Iterable,
    Literal,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
)
import sqlglot as sg

from odbc2deltalake.load_infos import (
    get_local_delta_value_and_count,
    retrieve_source_ts_cnt,
)
from odbc2deltalake.load_result import (
    AppendOnlyLoadResult,
    DeltaLoadResult,
    FullLoadResult,
    LoadResult,
    NoLoadResult,
)
from .utils import is_pydantic_2
from odbc2deltalake.sql_schema import is_string_type
from .utils import concat_seq
from odbc2deltalake.destination.destination import (
    Destination,
)
from odbc2deltalake.reader import DataSourceReader
from .query import sql_quote_name
from .metadata import (
    table_name_type,
    InformationSchemaColInfo,
)
import json
import sqlglot.expressions as ex
from .sql_glot_utils import table_from_tuple, union, count_limit_one
from .delta_logger import DeltaLogger
from .write_init import (
    WriteConfig,
    WriteConfigAndInfos,
    IS_DELETED_COL_NAME,
    IS_FULL_LOAD_COL_NAME,
    VALID_FROM_COL_NAME,
    DBDeltaPathConfigs,
)
from typing import Union

T = TypeVar("T")


def _source_convert(
    name: str,
    data_type: ex.DataType,
    *,
    table_alias: Union[str, None] = None,
    type_map: Optional[Mapping[str, ex.DataType]] = None,
):
    expr = ex.column(name, table_alias, quoted=True)

    data_type_str = (
        data_type
        if isinstance(data_type, str)
        else ex.DataType(this=data_type.this).sql("tsql").lower()
    )
    mapped_type = type_map.get(data_type_str) if type_map else None
    if mapped_type:
        expr = ex.cast(expr, mapped_type)
    if is_string_type(mapped_type or data_type):
        expr = ex.func("TRIM", expr)
    return expr


valid_from_expr = ex.cast(
    ex.func("GETUTCDATE", dialect="tsql"), ex.DataType(this="datetime2(6)")
).as_(VALID_FROM_COL_NAME, quoted=True)


def _get_cols_select(
    cols: Sequence[InformationSchemaColInfo],
    *,
    is_deleted: Union[bool, None] = None,
    is_full: Union[bool, None] = None,
    with_valid_from: bool = False,
    table_alias: Union[str, None] = None,
    system: Literal["source", "target"],
    data_type_map: Optional[Mapping[str, ex.DataType]] = None,
    get_target_name: Optional[Callable[[InformationSchemaColInfo], str]],
) -> Sequence[ex.Expression]:
    if get_target_name is None:
        get_target_name = lambda c: c.column_name

    return (
        [
            (
                _source_convert(
                    c.column_name,
                    c.data_type,
                    table_alias=table_alias,
                    type_map=data_type_map,
                ).as_(get_target_name(c), quoted=True)
                if system == "source"
                else ex.column(get_target_name(c), table_alias, quoted=True)
            )
            for c in cols
        ]
        + ([valid_from_expr] if with_valid_from else [])
        + (
            [
                ex.cast(ex.convert(int(is_deleted)), "bit").as_(
                    IS_DELETED_COL_NAME, quoted=True
                )
            ]
            if is_deleted is not None
            else []
        )
        + (
            [
                ex.cast(ex.convert(int(is_full)), "bit").as_(
                    IS_FULL_LOAD_COL_NAME, quoted=True
                )
            ]
            if is_full is not None
            else []
        )
    )


def _vacuum(source: DataSourceReader, dest: Destination):
    if source.local_delta_table_exists(dest):
        source.get_local_delta_ops(dest).vacuum()


def _transform_dt(dt: dict, dialect_src: str, dialect_trg: str):
    dt["data_type_src"] = dt["data_type"].sql(dialect_src)
    dt["data_type"] = dt["data_type"].sql(dialect_trg)
    return dt


def exec_write_db_to_delta(infos: WriteConfigAndInfos) -> LoadResult:
    write_config = infos.write_config
    cols = infos.col_infos
    pk_cols = infos.pk_cols
    destination = infos.destination
    source = infos.source
    delta_path = destination / "delta"
    dest_logger = infos.logger
    delta_col = infos.delta_col
    (destination / "meta").mkdir()
    (destination / "meta/schema.json").upload_str(
        json.dumps(
            [
                _transform_dt(
                    c.model_dump() if is_pydantic_2 else c.dict(),
                    infos.write_config.dialect,
                    source.query_dialect,
                )
                for c in cols
            ],
            indent=4,
        )
    )
    if source.local_delta_table_exists(
        destination / "delta_load" / DBDeltaPathConfigs.LATEST_PK_VERSION
    ):
        try:
            last_version_pk = source.get_local_delta_ops(
                destination / "delta_load" / DBDeltaPathConfigs.LATEST_PK_VERSION
            ).version()
        except Exception as e:
            import traceback

            dest_logger.warning(
                f"Could not get last version: {e}",
                error_trackback=traceback.format_exc(),
            )
            last_version_pk = None
    else:
        last_version_pk = None
    lock_file_path = destination / "meta/lock.txt"

    try:
        if (
            lock_file_path.exists()
            and (
                datetime.now(tz=timezone.utc) - lock_file_path.modified_time()
            ).total_seconds()
            > 60 * 60
        ):
            lock_file_path.remove()
        lock_file_path.upload_str("")
        if (
            not source.local_delta_table_exists(delta_path)
            or write_config.load_mode == "overwrite"
        ):
            delta_path.mkdir()
            load_result = do_full_load(infos=infos, mode="overwrite")
        elif write_config.load_mode == "append_inserts":
            if delta_col is None and len(pk_cols) == 1 and pk_cols[0].is_identity:
                delta_col = pk_cols[0]  # identity columns are usually increasing
                infos = dataclasses.replace(infos, delta_col=delta_col)
            assert (
                delta_col is not None
            ), "Must provide delta column for append_inserts load"
            load_result = do_append_inserts_load(infos)
        else:
            if (
                delta_col is None
                or len(pk_cols) == 0
                or write_config.load_mode == "force_full"
            ):
                load_result = do_full_load(
                    infos=infos,
                    mode="append",
                )
            else:
                load_result = do_delta_load(
                    infos=infos,
                    simple=write_config.load_mode
                    in ["simple_delta", "simple_delta_check"],
                    simple_check=write_config.load_mode == "simple_delta_check",
                )
        lock_file_path.remove()
        _vacuum(
            source, destination / "delta_load" / DBDeltaPathConfigs.LATEST_PK_VERSION
        )
        _vacuum(source, destination / "delta_load" / DBDeltaPathConfigs.DELTA_1_NAME)
        _vacuum(source, destination / "delta_load" / DBDeltaPathConfigs.DELTA_2_NAME)
        _vacuum(source, destination / "delta_load" / DBDeltaPathConfigs.PRIMARY_KEYS_TS)
        return load_result
    except Exception as e:
        # restore files
        if last_version_pk is not None:
            o = source.get_local_delta_ops(
                destination / "delta_load" / DBDeltaPathConfigs.LATEST_PK_VERSION
            )
            if o.version() > last_version_pk:
                o.restore(last_version_pk)
        import traceback

        dest_logger.error(
            "Error during load: {e}", error_trackback=traceback.format_exc()
        )
        raise e
    finally:
        if lock_file_path.exists():
            lock_file_path.remove()
        dest_logger.flush()


def _get_latest_pk_query(
    reader: DataSourceReader,
    destination: Destination,
    pks: Sequence[InformationSchemaColInfo],
    delta_col: InformationSchemaColInfo,
    write_config: WriteConfig,
    merge_delta=False,
    delta_load_value=None,
):
    reader.local_register_update_view(
        destination / f"delta_load/{DBDeltaPathConfigs.DELTA_1_NAME}",
        DBDeltaPathConfigs.DELTA_1_NAME,
    )

    reader.local_register_update_view(
        destination / f"delta_load/{DBDeltaPathConfigs.DELTA_2_NAME}",
        DBDeltaPathConfigs.DELTA_2_NAME,
    )
    if not merge_delta:
        reader.local_register_update_view(
            destination / f"delta_load/{DBDeltaPathConfigs.PRIMARY_KEYS_TS}",
            "primary_keys_ts_for_write",
        )

    def _nn(ls):
        return [l for l in ls if l is not None]

    return union(
        _nn(
            [
                ex.select(
                    *_get_cols_select(
                        cols=concat_seq(pks, [delta_col]),
                        table_alias="au",
                        system="target",
                        get_target_name=write_config.get_target_name,
                    )
                ).from_(table_from_tuple("delta_2", alias="au")),
                (
                    ex.select(
                        *_get_cols_select(
                            cols=concat_seq(pks, [delta_col]),
                            table_alias="d1",
                            system="target",
                            get_target_name=write_config.get_target_name,
                        )
                    )
                    .from_(ex.table_(DBDeltaPathConfigs.DELTA_1_NAME, alias="d1"))
                    .join(
                        ex.table_("delta_2", alias="au2"),
                        ex.and_(
                            *[
                                ex.column(
                                    write_config.get_target_name(c), "d1", quoted=True
                                ).eq(
                                    ex.column(
                                        write_config.get_target_name(c),
                                        "au2",
                                        quoted=True,
                                    )
                                )
                                for c in pks
                            ]
                        ),
                        join_type="anti",
                    )
                ),
                (
                    (
                        ex.select(
                            *_get_cols_select(
                                cols=concat_seq(pks, [delta_col]),
                                table_alias="cpk",
                                system="target",
                                get_target_name=write_config.get_target_name,
                            )
                        )
                        .from_(ex.table_("primary_keys_ts_for_write", alias="cpk"))
                        .where(
                            (
                                ex.column(
                                    write_config.get_target_name(delta_col), quoted=True
                                )
                                <= delta_load_value
                            )
                            if delta_load_value
                            else ex.convert(True)
                        )
                        .join(
                            ex.table_("delta_2", alias="au3"),
                            ex.and_(
                                *[
                                    ex.column(
                                        write_config.get_target_name(c),
                                        "cpk",
                                        quoted=True,
                                    ).eq(
                                        ex.column(
                                            write_config.get_target_name(c),
                                            "au3",
                                            quoted=True,
                                        )
                                    )
                                    for c in pks
                                ]
                            ),
                            join_type="anti",
                        )
                        .join(
                            ex.table_(DBDeltaPathConfigs.DELTA_1_NAME, alias="au4"),
                            ex.and_(
                                *[
                                    ex.column(
                                        write_config.get_target_name(c),
                                        "cpk",
                                        quoted=True,
                                    ).eq(
                                        ex.column(
                                            write_config.get_target_name(c),
                                            "au4",
                                            quoted=True,
                                        )
                                    )
                                    for c in pks
                                ]
                            ),
                            join_type="anti",
                        )
                    )
                    if not merge_delta
                    else None
                ),
            ]
        ),
        distinct=False,
    )


def write_latest_pk(
    reader: DataSourceReader,
    destination: Destination,
    pks: Sequence[InformationSchemaColInfo],
    delta_col: InformationSchemaColInfo,
    write_config: WriteConfig,
    merge_delta=False,
    delta_load_value=None,
):
    latest_pk_query = _get_latest_pk_query(
        reader,
        destination,
        pks,
        delta_col,
        write_config,
        merge_delta,
        delta_load_value=delta_load_value,
    )
    if merge_delta:
        reader.local_upsert_into(
            latest_pk_query,
            destination / "delta_load" / DBDeltaPathConfigs.LATEST_PK_VERSION,
            [write_config.get_target_name(pk) for pk in pks],
        )
    else:
        reader.local_execute_sql_to_delta(
            latest_pk_query,
            destination / "delta_load" / DBDeltaPathConfigs.LATEST_PK_VERSION,
            mode="overwrite",
            allow_schema_drift=True,
        )


def _temp_table(table: Union[table_name_type, ex.Query]):
    if isinstance(table, ex.Query):
        return "temp_" + str(abs(hash(table.sql("duckdb"))))

    def _clean(input_str: str):
        return "".join(ch for ch in input_str if ch.isalnum() and ch.isascii())

    if isinstance(table, str):
        return "temp_" + _clean(table)
    return "temp_" + "_".join((_clean(s) for s in table))


def do_delta_load(
    infos: WriteConfigAndInfos,
    simple=False,  # a simple delta load assumes that there are no deletes and no additional updates (eg, when soft-delete is implemented in source properly)
    simple_check=False,  # does a simple load and checks if the source and target counts match. If not, do a normal delta load on top
) -> LoadResult:
    delta_result = DeltaLoadResult()
    destination = infos.destination
    logger = infos.logger
    delta_col = infos.delta_col
    write_config = infos.write_config
    assert delta_col is not None, "Must have a delta_col for delta loads"
    reader = infos.source

    existing_col_infos = reader.get_local_delta_ops(
        destination / "delta"
    ).column_infos()

    existing_col_names = set((c.name.lower() for c in existing_col_infos))
    missing_cols = [
        write_config.get_target_name(c)
        for c in infos.col_infos
        if write_config.get_target_name(c).lower() not in existing_col_names
    ]
    if any(missing_cols) and infos.write_config.allow_schema_drift:
        logger.warning(f"New columns from source: {missing_cols}. Do a full load")
        return do_full_load(infos=infos, mode="append")

    last_pk_path = (
        destination / f"delta_load/{DBDeltaPathConfigs.LATEST_PK_VERSION}"
        if not simple
        else None
    )
    logger.info(
        f"Start { 'SIMPLE ' if simple else '' }Delta Load with Delta Column {delta_col.column_name} and pks: {', '.join((c.column_name for c in infos.pk_cols))}"
    )

    if last_pk_path and not reader.local_delta_table_exists(
        last_pk_path
    ):  # or do a full load?
        logger.warning("Primary keys missing, try to restore")
        try:
            from .write_utils.restore_pk import restore_last_pk

            restore_success = restore_last_pk(infos=infos)
        except Exception as e:
            logger.warning(f"Could not restore primary keys: {e}")
            restore_success = False
        if not restore_success:
            logger.warning("No primary keys found, do a full load")
            return do_full_load(infos=infos, mode="append")

    elif last_pk_path and not simple:
        cols = reader.get_local_delta_ops(last_pk_path).column_infos()
        cols = set((c.name.lower() for c in cols))
        pk_set = set((write_config.get_target_name(pk).lower() for pk in infos.pk_cols))
        if not cols.issuperset(pk_set):
            logger.warning(
                f"Primary keys do not match. Expected: {', '.join(pk_set)}, Found: {', '.join(cols)}. Do a full load"
            )
            return do_full_load(infos=infos, mode="append")

    old_pk_version = (
        reader.get_local_delta_ops(
            destination / "delta_load" / DBDeltaPathConfigs.LATEST_PK_VERSION
        ).version()
        if not simple
        else None
    )
    delta_path = destination / "delta"
    delta_load_value, current_count = get_local_delta_value_and_count(infos)
    delta_result.starting_local_state = delta_load_value, current_count
    source_delta, source_count = retrieve_source_ts_cnt(infos=infos)
    delta_result.starting_source_state = source_delta, source_count
    if (
        delta_load_value is not None
        and source_delta is not None
        and (delta_load_value, current_count) == (source_delta, source_count)
    ):
        logger.info("No updates, done")
        return NoLoadResult()

    if delta_load_value is None:
        logger.warning("No delta load value, do a full load")
        return do_full_load(
            infos=infos,
            mode="append",
        )

    if not simple:
        logger.info(
            f"Start delta step 1, get primary keys and timestamps. MAX({delta_col.column_name}): {delta_load_value}"
        )
        _retrieve_primary_key_data(infos=infos)
    else:
        logger.info(
            f"Start delta step 1, MAX({delta_col.column_name}): {delta_load_value}. Total RowCount: {source_count}"
        )
    criterion = _source_convert(
        delta_col.column_name,
        delta_col.data_type,
        table_alias="t",
        type_map=write_config.data_type_map,
    ) > ex.convert(delta_load_value)
    logger.info(
        f"Start delta step 2, load updates by {delta_col.column_name}",
        load="delta",
        sub_load="delta_1",
    )
    upds_sql = _get_update_sql(
        cols=infos.col_infos,
        criterion=criterion,
        query=infos.from_("t"),
        write_config=write_config,
    )
    _load_updates_to_delta(
        logger,
        reader,
        sql=upds_sql,
        delta_path=delta_path,
        delta_name="delta_1",
        write_config=write_config,
    )
    if not simple:
        assert old_pk_version is not None
        new_delta_load_value = _handle_additional_updates(
            infos=infos,
            old_pk_version=old_pk_version,
        )
        delta_load_value = new_delta_load_value or delta_load_value
        reader.local_register_update_view(delta_path, _temp_table(infos.table_or_query))

        logger.info("Start delta step 3.5, write deletes")
        do_deletes(
            reader=infos.source,
            destination=infos.destination,
            cols=infos.col_infos,
            pk_cols=infos.pk_cols,
            old_pk_version=old_pk_version,
            write_config=infos.write_config,
            delta_col=delta_col,
        )
        reader.local_register_update_view(delta_path, _temp_table(infos.table_or_query))
        logger.info("Start delta step 4, write meta for next delta load")
        write_latest_pk(
            reader,
            destination,
            infos.pk_cols,
            delta_col,
            write_config=write_config,
            delta_load_value=delta_load_value,
        )

        logger.info("Done delta load, do some last checks")
        target_count = _get_local_pk_count(infos)
        delta_result.dirty = source_count != target_count
        if source_count != target_count:
            logger.warning(
                f"Source and target count do not match. Source: {source_count}, Target: {target_count}. { 'Do a normal delta' if simple_check else ''}"
            )
            source_delta, source_count = retrieve_source_ts_cnt(infos=infos)
            delta_result.end_source_state = source_delta, source_count
            if delta_result.end_source_state != delta_result.starting_source_state:
                logger.warning(
                    f"Source state changed during load: {delta_result.starting_source_state} -> {delta_result.end_source_state}"
                )
            return delta_result

        else:
            logger.info(f"Source and target count match: {source_count}")
            return delta_result
    else:
        _write_delta2(infos, [], mode="overwrite")  # just to create the delta_2 table
        pk_ts = destination / "delta_load" / DBDeltaPathConfigs.PRIMARY_KEYS_TS
        if pk_ts.exists():
            pk_ts.remove(True)

        write_latest_pk(
            reader,
            destination,
            infos.pk_cols,
            delta_col,
            write_config=write_config,
            merge_delta=True,
        )
        target_count = _get_local_pk_count(infos)
        delta_result.dirty = source_count != target_count
        if source_count != target_count:
            logger.warning(
                f"Source and target count do not match. Source: {source_count}, Target: {target_count}. { 'Do a normal delta' if simple_check else ''}"
            )
            if simple_check:
                return do_delta_load(infos, simple=False)
            else:
                source_delta, source_count = retrieve_source_ts_cnt(infos=infos)
                delta_result.end_source_state = source_delta, source_count
                if delta_result.end_source_state != delta_result.starting_source_state:
                    logger.warning(
                        f"Source state changed during load: {delta_result.starting_source_state} -> {delta_result.end_source_state}"
                    )
                return delta_result
        else:
            logger.info(f"Source and target count match: {source_count}")
            return delta_result


def _get_local_pk_count(infos: WriteConfigAndInfos):
    reader = infos.source

    reader.local_register_update_view(
        infos.destination / "delta_load" / DBDeltaPathConfigs.LATEST_PK_VERSION,
        DBDeltaPathConfigs.LATEST_PK_VERSION,
    )
    return reader.local_execute_sql_to_py(
        sg.from_(DBDeltaPathConfigs.LATEST_PK_VERSION).select(
            ex.Count(this=ex.Star()).as_("cnt")
        )
    )[0]["cnt"]


def do_append_inserts_load(infos: WriteConfigAndInfos) -> AppendOnlyLoadResult:
    logger = infos.logger
    write_config = infos.write_config
    assert infos.delta_col is not None, "must have a delta col"
    logger.info(
        f"Start Append Only Load with Delta Column {infos.delta_col.column_name}"
    )
    delta_load_value, _ = get_local_delta_value_and_count(infos)

    criterion = (
        _source_convert(
            infos.delta_col.column_name,
            infos.delta_col.data_type,
            table_alias="t",
            type_map=write_config.data_type_map,
        )
        > ex.convert(delta_load_value)
        if delta_load_value
        else None
    )
    logger.info(" Start delta step 2, load updates by timestamp")
    _load_updates_to_delta(
        logger,
        infos.source,
        sql=_get_update_sql(
            cols=infos.col_infos,
            criterion=criterion,
            query=infos.from_("t"),
            write_config=write_config,
        ),
        delta_path=infos.destination / "delta",
        delta_name="delta_1",
        write_config=write_config,
    )

    logger.info("Done Append only load")
    return AppendOnlyLoadResult()


def do_deletes(
    reader: DataSourceReader,
    destination: Destination,
    # delta_table: DeltaTable,
    cols: Sequence[InformationSchemaColInfo],
    pk_cols: Sequence[InformationSchemaColInfo],
    delta_col: InformationSchemaColInfo,
    old_pk_version: int,
    write_config: WriteConfig,
):
    latest_pk_query = _get_latest_pk_query(
        reader,
        destination,
        pk_cols,
        delta_col=delta_col,
        write_config=write_config,
        merge_delta=False,
    )
    LAST_PK_VERSION = "LAST_PK_VERSION"
    reader.local_register_update_view(
        destination / f"delta_load/{ DBDeltaPathConfigs.LATEST_PK_VERSION}",
        LAST_PK_VERSION,
        version=old_pk_version,
    )
    delete_query = ex.except_(
        ex.select(
            *_get_cols_select(
                pk_cols,
                table_alias="lpk",
                system="target",
                get_target_name=write_config.get_target_name,
            )
        ).from_(table_from_tuple(LAST_PK_VERSION, alias="lpk")),
        ex.select(
            *_get_cols_select(
                pk_cols,
                table_alias="cpk",
                system="target",
                get_target_name=write_config.get_target_name,
            )
        ).from_(table_from_tuple("current_pk_version", alias="cpk")),
    ).with_("current_pk_version", as_=latest_pk_query)

    non_pk_cols = [c for c in cols if c not in pk_cols]
    non_pk_select = [
        ex.Null().as_(write_config.get_target_name(c), quoted=True) for c in non_pk_cols
    ]
    deletes_with_schema = union(
        [
            ex.select(
                *_get_cols_select(
                    pk_cols,
                    table_alias="d1",
                    system="target",
                    get_target_name=write_config.get_target_name,
                )
            )
            .select(
                *_get_cols_select(
                    non_pk_cols,
                    table_alias="d1",
                    system="target",
                    get_target_name=write_config.get_target_name,
                ),
                append=True,
            )
            .select(
                ex.AtTimeZone(
                    this=ex.CurrentTimestamp(),
                    zone=ex.Literal(this="UTC", is_string=True),
                ).as_(VALID_FROM_COL_NAME, quoted=True),
                ex.convert(True).as_(IS_DELETED_COL_NAME, quoted=True),
                ex.convert(False).as_(IS_FULL_LOAD_COL_NAME, quoted=True),
            )
            .from_(table_from_tuple("delta_1", alias="d1"))
            .where("1=0"),  # only used to get correct datatypes
            ex.select(
                ex.Column(this=ex.Star(), table=ex.Identifier(this="d", quoted=False))
            )
            .select(*non_pk_select, append=True)
            .select(
                ex.AtTimeZone(
                    this=ex.CurrentTimestamp(),
                    zone=ex.Literal(this="UTC", is_string=True),
                ).as_(VALID_FROM_COL_NAME, quoted=True),
                append=True,
            )
            .select(ex.convert(True).as_(IS_DELETED_COL_NAME, quoted=True), append=True)
            .select(
                ex.convert(False).as_(IS_FULL_LOAD_COL_NAME, quoted=True), append=True
            )
            .from_(table_from_tuple("deletes", alias="d")),
        ],
        distinct=False,
    ).with_("deletes", as_=delete_query)
    reader.local_register_view(deletes_with_schema, "deletes_with_schema")
    has_deletes = (
        reader.local_execute_sql_to_py(count_limit_one("deletes_with_schema"))[0]["cnt"]
        > 0
    )
    if has_deletes:
        reader.local_execute_sql_to_delta(
            sg.from_("deletes_with_schema").select("*"),
            destination / "delta",
            mode="append",
            allow_schema_drift=write_config.allow_schema_drift,
        )


def _retrieve_primary_key_data(
    infos: WriteConfigAndInfos,
):
    pk_ts_col_select = infos.from_("t").select(
        *_get_cols_select(
            is_full=None,
            is_deleted=None,
            cols=concat_seq(
                infos.pk_cols, [infos.delta_col] if infos.delta_col else []
            ),
            with_valid_from=False,
            data_type_map=infos.write_config.data_type_map,
            system="source",
            get_target_name=infos.write_config.get_target_name,
        )
    )
    pk_ts_reader_sql = pk_ts_col_select.sql(infos.write_config.dialect)

    pk_path = infos.destination / f"delta_load/{DBDeltaPathConfigs.PRIMARY_KEYS_TS}"
    infos.logger.info("Retrieve all PK/TS", sql=pk_ts_reader_sql)
    infos.source.source_write_sql_to_delta(
        sql=pk_ts_reader_sql,
        delta_path=pk_path,
        mode="overwrite",
        allow_schema_drift=True,
    )
    return pk_path


T = TypeVar("T")


def _list_to_chunks(input: Iterable[T], chunk_size: int):
    chunk: list[T] = list()
    for item in input:
        chunk.append(item)
        if len(chunk) >= chunk_size:
            yield chunk
            chunk = list()
    if len(chunk) > 0:
        yield chunk


def _write_delta2(
    infos: WriteConfigAndInfos, data: list[dict], mode: Literal["overwrite", "append"]
):
    write_config = infos.write_config
    from .query import sql_quote_value

    def _collate(c: InformationSchemaColInfo):
        if is_string_type(c.data_type):
            return "COLLATE Latin1_General_100_BIN "
        return ""

    delta_2_path = infos.destination / "delta_load" / DBDeltaPathConfigs.DELTA_2_NAME

    def full_sql(js: str):
        col_defs = ", ".join(
            [
                f"p{i} {p.data_type.sql(infos.write_config.dialect)}"
                for i, p in enumerate(infos.pk_cols)
            ]
        )

        selects = list(
            _get_cols_select(
                infos.col_infos,
                is_full=False,
                is_deleted=False,
                with_valid_from=True,
                table_alias="t",
                data_type_map=write_config.data_type_map,
                system="source",
                get_target_name=write_config.get_target_name,
            )
        )
        sql = infos.from_("t").select(*selects).sql(write_config.dialect)
        pk_map = ", ".join(
            [
                "p" + str(i) + " as " + sql_quote_name(write_config.get_target_name(c))
                for i, c in enumerate(infos.pk_cols)
            ]
        )
        return f"""{sql}
        inner join (SELECT {pk_map} FROM OPENJSON({sql_quote_value(js)}) with ({col_defs}) ) ttt
             on {' AND '.join([f't.{sql_quote_name(c.column_name)} {_collate(c)} = ttt.{sql_quote_name(write_config.get_target_name(c))}' for c in infos.pk_cols])}
        """

    if mode == "overwrite":
        infos.logger.info(
            "execute sql",
            load="delta",
            sub_load="delta_additional",
            sql=full_sql(f"[/* {len(data)} entries */]"),
        )
    sql = full_sql(json.dumps(data, default=str))
    if (
        len(sql) > 7000
    ):  ## oops, spark will not like this (actually the limit is 8000, but spark might use something on it's own)
        ch_split = len(data) // 2
        chunk_1 = data[:ch_split]
        chunk_2 = data[ch_split:]
        infos.source.source_write_sql_to_delta(
            full_sql(json.dumps(chunk_1, default=str)),
            delta_2_path,
            mode=mode,
            allow_schema_drift=write_config.allow_schema_drift,
        )
        infos.source.source_write_sql_to_delta(
            full_sql(json.dumps(chunk_2, default=str)),
            delta_2_path,
            mode="append",
            allow_schema_drift=write_config.allow_schema_drift,
        )
    else:
        infos.source.source_write_sql_to_delta(
            sql,
            delta_2_path,
            mode=mode,
            allow_schema_drift=write_config.allow_schema_drift,
        )


def _handle_additional_updates(
    infos: WriteConfigAndInfos,
    old_pk_version: int,
):
    """Handles updates that are not logical by their timestamp. This can happen on a restore from backup, for example."""
    folder = infos.destination
    delta_col = infos.delta_col
    pk_cols = infos.pk_cols
    cols = infos.col_infos
    reader = infos.source
    logger = infos.logger
    write_config = infos.write_config
    assert delta_col is not None, "Need a delta column"
    pk_ds_cols = concat_seq(pk_cols, [delta_col])
    reader.local_register_update_view(
        folder / f"delta_load/{ DBDeltaPathConfigs.PRIMARY_KEYS_TS}",
        DBDeltaPathConfigs.PRIMARY_KEYS_TS,
    )
    LAST_PK_VERSION = "LAST_PK_VERSION"
    reader.local_register_update_view(
        folder / f"delta_load/{ DBDeltaPathConfigs.LATEST_PK_VERSION}",
        LAST_PK_VERSION,
        version=old_pk_version,
    )

    def _local_view_for_updates():
        reader.local_register_view(
            ex.except_(
                ex.select(
                    *_get_cols_select(
                        cols=pk_ds_cols,
                        table_alias="pk",
                        system="target",
                        get_target_name=write_config.get_target_name,
                    )
                ).from_(ex.table_(DBDeltaPathConfigs.PRIMARY_KEYS_TS, alias="pk")),
                ex.select(
                    *_get_cols_select(
                        cols=pk_ds_cols,
                        table_alias="lpk",
                        system="target",
                        get_target_name=write_config.get_target_name,
                    )
                ).from_(table_from_tuple(LAST_PK_VERSION, alias="lpk")),
            ),
            "additional_updates",
        )

    try:
        _local_view_for_updates()
    except Exception as e:
        import traceback

        logger.warning(
            f"Could not create view for additional updates: {e}",
            error_trackback=traceback.format_exc(),
        )
        try:
            from .write_utils.restore_pk import restore_last_pk

            restore_success = restore_last_pk(infos=infos)
        except Exception as e:
            logger.warning(f"Could not restore primary keys: {e}")
            restore_success = False
        if not restore_success:
            logger.warning("No primary keys found, do a full load")
            do_full_load(infos=infos, mode="append")
            return
        else:
            _local_view_for_updates()
    sql_query = ex.except_(
        ex.select(
            *_get_cols_select(
                cols=pk_cols,
                table_alias="au",
                system="target",
                get_target_name=write_config.get_target_name,
            )
        ).from_(ex.table_("additional_updates", alias="au")),
        ex.select(
            *_get_cols_select(
                cols=pk_cols,
                table_alias="d1",
                system="target",
                get_target_name=write_config.get_target_name,
            )
        ).from_(table_from_tuple("delta_1", alias="d1")),
    )
    reader.local_register_view(sql_query, "real_additional_updates")
    update_count: int = reader.local_execute_sql_to_py(
        sg.from_("real_additional_updates").select(ex.Count(this=ex.Star()).as_("cnt"))
    )[0]["cnt"]

    jsd = reader.local_execute_sql_to_py(
        sg.from_("real_additional_updates").select(
            *[
                ex.column(write_config.get_target_name(c)).as_(
                    "p" + str(i), quoted=False
                )
                for i, c in enumerate(pk_cols)
            ]
        )
    )

    if update_count == 0:
        _write_delta2(infos, [], mode="overwrite")
    elif (
        (update_count > 1000) or write_config.no_complex_entries_load
    ):  # many updates. get the smallest timestamp and do "normal" delta, even if there are too many records then
        _write_delta2(infos, [], mode="overwrite")  # still need to create delta_2_path
        logger.warning(
            f"Start delta step 3, load {update_count} strange updates via normal delta load"
        )
        delta_load_value = reader.local_execute_sql_to_py(
            ex.select(
                ex.func(
                    "MIN",
                    ex.column(write_config.get_target_name(delta_col), quoted=True),
                ).as_("min_ts")
            ).from_(ex.table_("additional_updates", alias="rau"))
        )[0]["min_ts"]
        criterion = _source_convert(
            delta_col.column_name,
            delta_col.data_type,
            table_alias="t",
            type_map=write_config.data_type_map,
        ) > ex.convert(delta_load_value)
        logger.info("Start delta step 2, load updates by timestamp")
        upds_sql = _get_update_sql(
            cols=cols,
            criterion=criterion,
            query=infos.from_("t"),
            write_config=write_config,
        )
        logger.info(
            "execute sql", load="delta", sub_load="delta_1_additional", sql=upds_sql
        )
        _load_updates_to_delta(
            logger,
            reader,
            sql=upds_sql,
            delta_path=infos.destination / "delta",
            delta_name="delta_1",
            write_config=write_config,
        )
        return delta_load_value
    else:
        # we don't want to overshoot 8000 chars here because of spark. we estimate how much space in json a record of pk's will take

        char_size_pks = sum(
            [
                5
                + (
                    10
                    if p.data_type.this
                    in [
                        ex.DataType.Type.BOOLEAN,
                        *ex.DataType.NUMERIC_TYPES,
                    ]
                    else 40
                )
                for p in pk_cols
            ]
        )
        batch_size = max(10, int(7000 / char_size_pks))

        logger.warning(
            f"Start delta step 3, load {update_count} strange updates via batches of size {batch_size}"
        )
        first = True
        for chunk in _list_to_chunks(jsd, batch_size):
            _write_delta2(infos, chunk, mode="overwrite" if first else "append")
            first = False
        reader.local_register_update_view(
            infos.destination / "delta_load" / DBDeltaPathConfigs.DELTA_2_NAME,
            "delta_2",
        )
        reader.local_execute_sql_to_delta(
            sg.from_("delta_2").select(ex.Star()),
            infos.destination / "delta",
            mode="append",
            allow_schema_drift=True,
        )
        return None


def _get_update_sql(
    cols: Sequence[InformationSchemaColInfo],
    criterion: Union[Sequence[ex.Expression], ex.Expression, None],
    query: ex.Select,
    write_config: WriteConfig,
):
    if isinstance(criterion, ex.Expression):
        criterion = [criterion]
    delta_sql = (
        query.select(
            *_get_cols_select(
                cols,
                is_full=False,
                is_deleted=False,
                with_valid_from=True,
                table_alias="t",
                data_type_map=write_config.data_type_map,
                system="source",
                get_target_name=write_config.get_target_name,
            )
        )
        .where(
            *(
                criterion
                if criterion is not None and not isinstance(criterion, str)
                else []
            ),
            dialect=write_config.dialect,
        )
        .sql(write_config.dialect)
    )
    return delta_sql


def _load_updates_to_delta(
    logger: DeltaLogger,
    reader: DataSourceReader,
    delta_path: Destination,
    sql: Union[str, ex.Query],
    delta_name: str,
    write_config: WriteConfig,
):
    if isinstance(sql, ex.Query):
        sql = sql.sql(write_config.dialect)

    delta_name_path = delta_path.parent / f"delta_load/{delta_name}"
    logger.info("Executing sql", load="delta", sub_load=delta_name, sql=sql)
    reader.source_write_sql_to_delta(
        sql,
        delta_name_path,
        mode="overwrite",
        allow_schema_drift=write_config.allow_schema_drift,
    )
    reader.local_register_update_view(delta_name_path, delta_name)
    count = reader.local_execute_sql_to_py(count_limit_one(delta_name))[0]["cnt"]
    if count == 0:
        return
    reader.local_execute_sql_to_delta(
        sg.from_(delta_name).select(ex.Star()),
        delta_path,
        mode="append",
        allow_schema_drift=write_config.allow_schema_drift,
    )


def do_full_load(
    infos: WriteConfigAndInfos, mode: Literal["overwrite", "append"]
) -> FullLoadResult:
    logger = infos.logger
    write_config = infos.write_config
    delta_path = infos.destination / "delta"
    reader = infos.source

    logger.info("Start Full Load")
    sql = (
        infos.from_("t")
        .select(
            *_get_cols_select(
                is_deleted=False,
                is_full=True,
                cols=infos.col_infos,
                with_valid_from=True,
                data_type_map=write_config.data_type_map,
                system="source",
                get_target_name=write_config.get_target_name,
            )
        )
        .sql(write_config.dialect)
    )
    logger.info("executing sql", sql=sql, load="full")
    reader.source_write_sql_to_delta(
        sql, delta_path, mode=mode, allow_schema_drift=write_config.allow_schema_drift
    )
    if infos.delta_col is None:
        logger.info("Full Load done")
        return FullLoadResult()
    logger.info(" Full Load done, write meta for delta load")

    reader.local_register_update_view(delta_path, _temp_table(infos.table_or_query))
    (delta_path.parent / "delta_load").mkdir()
    ident = ex.to_identifier(_temp_table(infos.table_or_query))
    query = (
        sg.from_(ident)
        .select(
            *(
                [
                    ex.column(write_config.get_target_name(pk), quoted=True)
                    for pk in infos.pk_cols
                ]
                + (
                    [
                        ex.column(
                            write_config.get_target_name(infos.delta_col), quoted=True
                        )
                    ]
                    if infos.delta_col
                    else []
                )
            )
        )
        .where(
            ex.column(VALID_FROM_COL_NAME, quoted=True).eq(
                sg.from_(ident)
                .select(ex.func("MAX", ex.column(VALID_FROM_COL_NAME, quoted=True)))
                .where(ex.column(IS_FULL_LOAD_COL_NAME, quoted=True))
                .subquery()
            )
        )
    )
    reader.local_execute_sql_to_delta(
        query,
        delta_path.parent / "delta_load" / DBDeltaPathConfigs.LATEST_PK_VERSION,
        mode="overwrite",
        allow_schema_drift=True,
    )
    return FullLoadResult()



from typing import Union
import sqlglot.expressions as ex

table_name_type = Union[str, tuple[str, str]]


def is_string_type(type: Union[ex.DataType, ex.DataType.Type]):
    if isinstance(type, ex.DataType):
        return type.is_string or type.this in ex.DataType.TEXT_TYPES
    return type in ex.DataType.TEXT_TYPES



from abc import ABC, abstractmethod
from datetime import datetime
from typing_extensions import Literal, Self
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from deltalake import DeltaTable


class Destination(ABC):
    @abstractmethod
    def mkdir(self):
        pass

    @abstractmethod
    def __truediv__(self, other: str) -> Self:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def exists(self) -> bool:
        pass

    @abstractmethod
    def upload_str(self, data: str):
        pass

    @abstractmethod
    def modified_time(self) -> datetime:
        pass

    @abstractmethod
    def remove(self, recurse: bool = False):
        pass

    @property
    @abstractmethod
    def parent(self) -> Self:
        pass

    @abstractmethod
    def as_path_options(
        self, flavor: Literal["fsspec", "object_store"]
    ) -> tuple[str, Optional[dict[str, str]]]:
        pass

    @abstractmethod
    def as_delta_table(self) -> "DeltaTable":
        pass



from typing import TypeVar, Sequence

T = TypeVar("T")


def concat_seq(*args: Sequence[T]) -> Sequence[T]:
    if len(args) == 1:
        return args[0]
    r = list()
    for a in args:
        r = r + a  # type: ignore

    return r


def _is_pydantic_2() -> bool:
    import pydantic

    try:
        # __version__ was added with Pydantic 2 so we know if this errors the version is < 2.
        # Still check the version as a fail safe incase __version__ gets added to verion 1.
        if int(pydantic.__version__[:1]) >= 2:  # type: ignore[attr-defined]
            return True

        # Raise an AttributeError to match the AttributeError on __version__ because in either
        # case we need to get to the same place.
        raise AttributeError  # pragma: no cover
    except AttributeError:  # pragma: no cover
        return False


is_pydantic_2 = _is_pydantic_2()



from typing import Literal, Union

from odbc2deltalake.query import sql_quote_name
from .reader import DataSourceReader
from pydantic import BaseModel
import sqlglot
import sqlglot.expressions as ex
from .utils import is_pydantic_2

if is_pydantic_2:
    from pydantic import ConfigDict

table_name_type = Union[str, tuple[str, str], tuple[str, str, str]]


def get_primary_keys(
    reader: DataSourceReader, table_name: table_name_type, *, dialect: str
) -> list[str]:
    if isinstance(table_name, str):
        table_name = ("dbo", table_name)
    real_table_name = table_name[1] if len(table_name) == 2 else table_name[2]
    real_schema = table_name[0] if len(table_name) == 2 else table_name[1]
    real_db = table_name[0] if len(table_name) == 3 else None
    quoted_db = sql_quote_name(real_db) + "." if real_db else ""

    query = sqlglot.parse_one(
        f"""SELECT ccu.COLUMN_NAME
    FROM {quoted_db}INFORMATION_SCHEMA.TABLE_CONSTRAINTS tc WITH(NOLOCK)
        JOIN {quoted_db}INFORMATION_SCHEMA.CONSTRAINT_COLUMN_USAGE ccu WITH(NOLOCK) ON tc.CONSTRAINT_NAME = ccu.Constraint_name
    WHERE tc.CONSTRAINT_TYPE = 'Primary Key'""",
        dialect=dialect,
    )
    assert isinstance(query, ex.Select)
    query = query.where(
        ex.column("TABLE_NAME", "ccu")
        .eq(ex.convert(real_table_name))
        .and_(ex.column("TABLE_SCHEMA", "ccu").eq(ex.convert(real_schema)))
    )
    full_query = query.sql(dialect)
    return [d["COLUMN_NAME"] for d in reader.source_sql_to_py(full_query)]


class FieldWithType(BaseModel):
    name: str
    type: str
    max_str_length: Union[int, None] = None


class InformationSchemaColInfo(BaseModel):
    if is_pydantic_2:
        model_config = ConfigDict(arbitrary_types_allowed=True)
    if not is_pydantic_2:

        class Config:
            arbitrary_types_allowed = True

    column_name: str

    data_type: ex.DataType
    column_default: Union[str, None] = None
    is_nullable: bool = True
    generated_always_type_desc: Union[
        Literal["NOT_APPLICABLE", "AS_ROW_START", "AS_ROW_END"], None
    ] = "NOT_APPLICABLE"
    is_identity: bool = False


def _get_table_cols(
    reader: DataSourceReader, table_name: table_name_type, *, dialect: str
):
    if isinstance(table_name, str):
        table_name = ("dbo", table_name)
    real_table_name = table_name[1] if len(table_name) == 2 else table_name[2]
    real_schema = table_name[0] if len(table_name) == 2 else table_name[1]
    real_db = table_name[0] if len(table_name) == 3 else None
    quoted_db = sql_quote_name(real_db) + "." if real_db else ""

    query = sqlglot.parse_one(
        f""" SELECT  ccu.column_name, ccu.column_default,
		cast(case when ccu.IS_NULLABLE='YES' THEN 1 ELSE 0 END as bit) as is_nullable,
		data_type,
		character_maximum_length,
		numeric_precision,
		numeric_scale,
		datetime_precision,
        ci.generated_always_type_desc,
        coalesce(ci.is_identity, convert(bit, 0)) as is_identity FROM {quoted_db}INFORMATION_SCHEMA.COLUMNS ccu
        left join (
			
SELECT sc.name as schema_name, t.name as table_name, c.name as col_name, c.generated_always_type_desc, c.is_identity as is_identity FROM {quoted_db}sys.columns c 
	inner join {quoted_db}sys.tables t on t.object_id=c.object_id
	inner join {quoted_db}sys.schemas sc on sc.schema_id=t.schema_id
		) ci on ci.schema_name=ccu.TABLE_SCHEMA and ci.table_name=ccu.TABLE_NAME and ci.col_name=ccu.COLUMN_NAME
		 """,
        dialect=dialect,
    )
    assert isinstance(query, ex.Select)
    full_query = query.where(
        ex.column("TABLE_NAME", "ccu")
        .eq(ex.convert(real_table_name))
        .and_(ex.column("TABLE_SCHEMA", "ccu").eq(ex.convert(real_schema)))
    ).sql(dialect)
    dicts = reader.source_sql_to_py(full_query)
    for d in dicts:
        dt = d.pop("data_type")
        max_len = d.pop("character_maximum_length")
        numeric_precision = d.pop("numeric_precision")
        numeric_scale = d.pop("numeric_scale")
        datetime_precision = d.pop("datetime_precision")
        if dt in ["xml"]:
            max_len = None
        if dt.lower() not in ["numeric", "decimal"]:
            numeric_precision = None
            numeric_scale = None
        dt_str = dt
        if max_len is not None:
            dt_str += f"({max_len})" if max_len != -1 else "(MAX)"
        elif numeric_precision is not None:
            dt_str += f"({numeric_precision},{numeric_scale})"
        elif datetime_precision and dt.lower() not in ["datetime", "date"]:
            dt_str += f"({datetime_precision})"

        d["data_type"] = ex.DataType.build(dt_str, dialect=dialect)
    return [InformationSchemaColInfo(**d) for d in dicts]


def _get_query_cols_first_result_set(
    reader: DataSourceReader, query: ex.Query, *, dialect: str
):
    sql = (
        "EXEC sp_describe_first_result_set @tsql=N'"
        + query.sql(dialect).replace("'", "''")
        + "'"
    )

    dicts = reader.source_sql_to_py(sql)

    for d in dicts:
        sys_type_name = d["system_type_name"]

        yield InformationSchemaColInfo(
            column_name=d["name"],
            data_type=ex.DataType.build(sys_type_name, dialect=dialect),
            column_default=None,
            is_identity=d["is_identity_column"],
            generated_always_type_desc="NOT_APPLICABLE",
            is_nullable=d["is_nullable"],
        )


def get_columns(
    reader: DataSourceReader,
    table_or_query: Union[table_name_type, ex.Query],
    *,
    dialect: str,
) -> list[InformationSchemaColInfo]:
    if isinstance(table_or_query, ex.Query):
        if not reader.supports_proc_exec or dialect != "tsql":
            return reader.source_schema_limit_one(
                table_or_query
            )  # get result metadata by just executing the query with limit 0
        else:
            return list(
                _get_query_cols_first_result_set(
                    reader, table_or_query, dialect=dialect
                )
            )

    else:
        return _get_table_cols(reader, table_or_query, dialect=dialect)


def get_compatibility_level(reader: DataSourceReader) -> int:
    return reader.source_sql_to_py(
        "SELECT compatibility_level FROM sys.databases WHERE name =DB_NAME()"
    )[0]["compatibility_level"]
